# tested with Python 3.5, july-2017 (1.12)
# tested with Python 2.4, september-2007 (1.12)
# tested with Python 2.5, december-2008 (1.12_1)
from __future__ import division, print_function, absolute_import
from builtins import range
from builtins import next
from builtins import zip
from past.builtins import cmp

import sys  # noqa
import unittest
import random
import string
import pickle
from io import BytesIO

from pyrocko import avl


def verify_empty(tree):
    return tree.verify() == 1 and len(tree) == 0


def verify_len(tree, size):
    return tree.verify() == 1 and len(tree) == size


def gen_ints(lo, hi):
    for i in range(lo, hi):
        yield i


# call with gcd(step,modulo)=1
def gen_ints_perm(step, modulo):
    n = random.randint(0, modulo-1)
    for i in range(modulo):
        val = n
        n = (val+step) % modulo
        yield val


def gen_pairs(lo, hi):
    for i in range(lo, hi):
        yield (i, 0)


def gen_iter(a, b):
    return zip(a, b)


def range_tree(lo, hi, compare=None):
    t = avl.new(compare=compare)
    for i in gen_ints(lo, hi):
        t.insert(i)
    return t


def random_int_tree(lo, hi, size=1000, compare=None):
    t = avl.new(compare=compare)
    for i in gen_ints(0, size):
        t.insert(random.randint(lo, hi-1))
    return t


def random_pair_tree(lo, hi, size=1000, compare=None):
    t = avl.new(compare=compare)
    for i in gen_ints(0, size):
        t.insert((random.randint(lo, hi-1), 0))
    return t


# return 1 if tree1 and tree2 hold the same set of keys
def equal_tree(tree1, tree2, compare=cmp):
    if len(tree1) != len(tree2):
        return 0
    i1 = iter(tree1)
    i2 = iter(tree2)
    try:
        while compare(next(i1), next(i2)) is 0:
            pass
        return 0
    except StopIteration:
        return 1


# sort tuples by decreasing order of first component
def pair_compare(t, u):
    if t[0] < u[0]:
        return 1
    if t[0] == u[0]:
        return 0
    return -1


# ----
# Test insertions
# ----
class Test_avl_ins(unittest.TestCase):

    def setUp(self):
        self.size = 1000
        self.list = list(range(self.size))
        self._times = 7

    # test avl.new
    def testnew_basic(self):
        for i in range(self._times):
            random.shuffle(self.list)
            t = avl.new(self.list, unique=0)
            self.assertTrue(t.verify() == 1 and len(t) == self.size)
            for n in self.list:
                self.assertTrue(n in t)
        t = None

    # test avl.new unique=0
    def testnew_basic_dupes(self):
        size3 = self.size * 3
        for i in range(self._times):
            random.shuffle(self.list)
            list3 = self.list * 3
            t = avl.new(list3, unique=0)
            self.assertTrue(verify_len(t, size3))
            for n in self.list:
                self.assertTrue(n in t)
        t = None

    # test avl.new unique=1
    def testnew_basic_unique(self):
        for i in range(self._times):
            random.shuffle(self.list)
            list3 = self.list * 3
            t = avl.new(list3, unique=1)
            self.assertTrue(verify_len(t, self.size))
            for n in self.list:
                self.assertTrue(n in t)
        t = None

    # test avl.new with compare function
    def testnew_compare_basic(self):
        random.shuffle(self.list)
        pairs = [(i, 0) for i in self.list]
        for i in range(self._times):
            t = avl.new(pairs, compare=pair_compare)
            self.assertTrue(verify_len(t, self.size))
            for p in pairs:
                self.assertTrue(p in t)

    # test avl.new with compare function, unique=0
    def testnew_compare_dupes(self):
        random.shuffle(self.list)
        pairs = [(i, 0) for i in self.list]
        pairs3 = pairs * 3
        size3 = self.size * 3
        for i in range(self._times):
            t = avl.new(pairs3, compare=pair_compare, unique=0)
            self.assertTrue(verify_len(t, size3))
            for p in pairs:
                self.assertTrue(p in t)

    # test avl.new with compare function, unique=1
    def testnew_compare_unique(self):
        random.shuffle(self.list)
        tuples = [(i, 0) for i in self.list]
        tuples3 = tuples * 3
        for i in range(self._times):
            t = avl.new(tuples3, compare=pair_compare, unique=1)
            self.assertTrue(verify_len(t, self.size))
            for u in tuples:
                self.assertTrue(u in t)

    # comparison failure
    def testins_bad(self):
        t = avl.new([(5, 6), (2, 8), (7, -1), (-1, 1), (2, -3)],
                    compare=pair_compare)
        for i in range(5):
            self.assertRaises(TypeError, t.insert, 66)
            self.assertRaises(TypeError, t.has_key, 66)
            self.assertFalse(66 in t)
            t.insert((3, 3))
            self.assertTrue(t.verify() == 1)

    # test ins and clear
    def testins_big(self):
        big_size = 20000
        for i in range(self._times):
            t = random_int_tree(-big_size, big_size, big_size)
            self.assertTrue(verify_len(t, big_size))
            t.clear()
            self.assertTrue(verify_empty(t))

    def tearDown(self):
        self.list = None


# ----
# Test lookup
# ----
class Test_avl_lookup(unittest.TestCase):

    def setUp(self):
        self.n = 5000
        self.t = range_tree(0, self.n)
        self.bad = list(range(-500, 0)) + list(range(self.n, self.n+500))

    def testlookup_get(self):
        for i in self.bad:
            self.assertRaises(LookupError, self.t.lookup, i)
        for i in gen_ints(0, self.n):
            self.assertTrue(self.t.lookup(i) == i)

    def testlookup_contains(self):
        for i in self.bad:
            self.assertFalse(i in self.t)
        for i in gen_ints(0, self.n):
            self.assertTrue(i in self.t)

    def testlookup_span_one(self):
        for i in gen_ints(0, self.n):
            a, b = self.t.span(i)
            self.assertTrue(b-a == 1)

    def testlookup_span_two(self):
        t = self.t
        for i in gen_ints(0, self.n):
            j = random.randint(i, self.n-1)
            self.assertTrue(t.span(i, j) == (i, j+1))
            self.assertTrue(t.span(j, i) == (i, j+1))
            self.assertTrue(t.span(i, self.n) == (i, self.n))
            self.assertTrue(t.span(i, self.n*2) == (i, self.n))
            self.assertTrue(t.span(-j, i) == (0, i+1))

    def tearDown(self):
        self.t = None
        self.bad = None


# ----
# Test iterations
# ----
class Test_avl_iter(unittest.TestCase):

    def setUp(self):
        self.n = 1000
        self.t = range_tree(0, self.n)
        self.orig = list(range(self.n))

    def testiter_forloop(self):
        list = self.orig[:]
        for i in range(5):
            random.shuffle(list)
            for k in avl.new(list):
                self.assertTrue(k == self.orig[k])

    def testiter_forward(self):
        j = self.t.iter()
        for k in gen_ints(0, self.n):
            self.assertTrue(next(j) == k and j.index() == k and j.cur() == k)
        self.assertRaises(StopIteration, j.__next__)
        self.assertRaises(avl.Error, j.cur)
        self.assertTrue(j.index() == self.n)

    def testiter_backward(self):
        j = self.t.iter(1)
        for k in gen_ints(1, self.n+1):
            self.assertTrue(j.prev()+k == self.n
                            and j.index()+k == self.n
                            and j.cur()+k == self.n)
        self.assertRaises(StopIteration, j.prev)
        self.assertRaises(avl.Error, j.cur)
        self.assertTrue(j.index() == -1)

    def testiter_basic(self):
        t = avl.new()
        j = iter(t)     # before first
        k = t.iter(1)   # after last
        self.assertRaises(StopIteration, j.__next__)
        self.assertRaises(StopIteration, j.prev)
        self.assertRaises(StopIteration, k.__next__)
        self.assertRaises(StopIteration, k.prev)
        t.insert('bb')
        self.assertRaises(StopIteration, j.prev)
        self.assertRaises(StopIteration, k.__next__)
        self.assertTrue(next(j) == 'bb')
        self.assertTrue(k.prev() == 'bb')
        self.assertRaises(StopIteration, j.__next__)
        self.assertRaises(StopIteration, k.prev)
        self.assertTrue(j.prev() == 'bb')
        self.assertTrue(next(k) == 'bb')
        self.assertTrue(j.cur() == 'bb' and k.cur() == 'bb')
        t.insert('aa')
        self.assertTrue(j.prev() == 'aa')
        t.insert('cc')
        self.assertTrue(next(k) == 'cc')

    def testiter_remove(self):
        for start in range(1, self.n+1):
            u = avl.new(self.t)
            self.assertTrue(u.verify() == 1)
            j = iter(u)
            for i in range(start):
                next(j)
            index = j.index()
            self.assertTrue(index == start-1)
            while index < len(u):
                j.remove()
                # self.assertTrue(u.verify() == 1)
                self.assertTrue(j.index() == index)
            self.assertRaises(avl.Error, j.remove)
            self.assertTrue(u.verify() == 1)

    def tearDown(self):
        self.t.clear()
        self.t = None
        self.orig = None


# ----
# Test duplication
# ----
class Test_avl_dup(unittest.TestCase):

    def setUp(self):
        self.n = 5000

    def testdup_basic(self):
        for i in range(10):
            t = random_int_tree(0, self.n, size=self.n)
            u = avl.new(t)
            v = t[:]
            self.assertTrue(u.verify() == 1)
            self.assertTrue(equal_tree(t, u))
            self.assertTrue(v.verify() == 1)
            self.assertTrue(equal_tree(t, v))
        t = avl.new()
        u = avl.new(t)
        self.assertTrue(verify_empty(u))

    def testdup_compare(self):
        for i in range(5):
            t = random_pair_tree(0, self.n, size=self.n, compare=pair_compare)
            u = avl.new(t)
            v = t[:]
            self.assertTrue(u.verify() == 1)
            self.assertTrue(equal_tree(t, u, pair_compare))
            self.assertTrue(v.verify() == 1)
            self.assertTrue(equal_tree(t, v, pair_compare))
        t = avl.new(compare=pair_compare)
        u = avl.new(t)
        self.assertTrue(verify_empty(u))


# ----
# Test deletions
# ----
class Test_avl_del(unittest.TestCase):

    def setUp(self):
        pass

    # empty tree with unique keys
    def testdel_basic(self):
        t = avl.new()
        t.remove(1)
        t.remove(-2)
        self.assertTrue(verify_empty(t))
        t = range_tree(-2000, +2000)
        self.assertTrue(t.verify() == 1)
        n = len(t)
        # no-op
        others = list(range(-2100, -2000)) + list(range(2000, 2100))
        random.shuffle(others)
        for i in others:
            t.remove(i)
        self.assertTrue(verify_len(t, n))
        others = None
        # empty trees
        lst = list(range(-2000, 2000))
        for i in range(10):
            random.shuffle(lst)
            u = avl.new(t)
            for k in lst:
                u.remove(k)
                self.assertFalse(k in u)
            self.assertTrue(verify_empty(u))

    # empty tree with duplicates
    def testdel_lessbasic(self):
        n = 1000
        t = avl.new()
        for i in range(3):
            for k in gen_ints(0, n):
                t.insert(k)
        self.assertTrue(t.verify() == 1)
        # params for sequence
        modulo = n
        step = 37
        for i in range(10):
            u = avl.new(t)
            for k in gen_ints_perm(step, modulo):
                while k in u:
                    u.remove(k)
                self.assertFalse(k in u)
            self.assertTrue(u.verify() == 1)
            self.assertTrue(len(u) == 0, 'len='+str(len(u))+' '+str(u))

    def testdel_bigunique(self):
        t = range_tree(-10000, 10000)
        for i in gen_ints_perm(10007, 20000):
            j = i - 10000
            self.assertTrue(j in t)
            t.remove(j)
            self.assertFalse(j in t)

    def testdel_one(self):
        n = 4000
        t = random_int_tree(0, n, n)
        for i in gen_ints_perm(1993, n):
            e = t[i]
            a1, a2 = t.span(e)
            t.remove(e)
            self.assertTrue(t.verify() == 1)
            b1, b2 = t.span(e)
            self.assertTrue(a2-a1-1 == b2-b1)
            t.insert(e)


# ----
# Test sequence support
# ----
class Test_avl_sequence(unittest.TestCase):

    def geti(self, t, i):
        return t[i]

    def testseq_basic(self):
        t = avl.new()
        for step, modulo in [(491, 2000), (313, 10000)]:
            for i in gen_ints_perm(step, modulo):
                t.insert(i)

            for i in gen_ints(0, modulo):
                self.assertTrue(t.index(i) == i)
                self.assertTrue(t[i] == i)
                self.assertTrue(t[-i-1] == t[modulo-i-1])

            for i in gen_ints(-100, 0):
                self.assertTrue(t.index(i) == -1)
            for i in gen_ints(modulo, modulo+100):
                self.assertTrue(t.index(i) == -1)
            t.clear()
        self.assertRaises(IndexError, self.geti, t, 11)
        self.assertRaises(IndexError, self.geti, t, -5)

    def testseq_index(self):
        t = avl.new()
        n = 5000
        for r in range(3):
            for i in range(3):
                for k in gen_ints_perm(11, n):
                    t.insert(k)
            self.assertTrue(verify_len(t, n*3))
            for i in gen_ints(0, n):
                self.assertTrue(t.index(i) == i*3)
                a, b = t.span(i)
                self.assertTrue((a == i*3) and (b-a == 3))
            t.clear()

    def testseq_insert(self):
        n = 5000
        t = avl.new()
        self.assertRaises(IndexError, t.insert, 'out', 1)
        self.assertRaises(IndexError, t.insert, 'out', -13)
        for i in gen_ints_perm(373, n):
            t.insert(2*i)
        self.assertTrue(t.verify() == 1)
        for i in range(3):
            u = avl.new(t)
            for k in gen_ints(0, n):
                odd = 2*k+1
                u.insert(odd, odd)
                self.assertTrue(u[odd] == odd)
        self.assertTrue(u.verify() == 1)
        self.assertTrue(len(u) == 2*n)
        for i in range(100):
            u.append(2*n)
        for i in range(100):
            u.insert(0, 0)
        self.assertTrue(u.verify() == 1)
        list = None  # noqa

    def testseq_remove(self):
        n = 5000
        t = range_tree(0, n)
        self.assertRaises(IndexError, t.remove_at, -n-random.randint(0, 100))
        for i in gen_ints_perm(29, n):
            t.remove_at(i)
            self.assertFalse(i in t)
            t.insert(i, i)
        self.assertTrue(verify_len(t, n))
        # empty the tree
        for i in gen_ints(0, n):
            k = random.randint(0, len(t)-1)
            e = t[k]
            t.remove_at(k)
            self.assertFalse(e in t or t.index(e) != -1)
        self.assertTrue(verify_empty(t))
        self.assertRaises(IndexError, t.remove_at, 0)

    def testseq_spaninsert(self):
        n = 5000
        t = avl.new(compare=lambda x, y: cmp(y, x))
        for i in range(2):
            for i in gen_ints(0, 3*n):
                e = random.randint(-n, n)
                a1, a2 = t.span(e)
                t.insert(e, a2)
                self.assertTrue(t.span(e) == (a1, a2+1))
            self.assertTrue(verify_len(t, 3*n))
            t.clear()

    def testseq_removedupes(self):
        n = 2500
        steps = [501, 7, 1003, 1863]
        repeats = len(steps)
        t = avl.new()
        for step in steps:
            for k in gen_ints_perm(step, n):
                t.insert(k)
        for k in gen_ints_perm(1209, n):
            a, b = t.span(k)
            self.assertTrue(b-a == repeats)
            self.assertTrue(a == 0 or t[a-1] < k)
            self.assertTrue(b == len(t) or t[b] > k)
            for i in range(repeats):
                t.remove_at(a)
            self.assertTrue(t.span(k) == (a, a))
        self.assertTrue(verify_empty(t))

    def testseq_slicedup(self):
        n = 3000
        for i in range(3):
            t = random_int_tree(0, n, 4*n)
            self.assertTrue(equal_tree(t[:], t))

    def testseq_sliceempty(self):
        t = random_int_tree(0, 500, size=1000)
        lim = len(t)
        for i in range(100):
            a = random.randint(0, lim)
            b = random.randint(0, a)
            self.assertTrue(verify_empty(t[a:b]))
            a = random.randint(1, lim)
            b = random.randint(1, a)
            self.assertTrue(verify_empty(t[-b:-a]))

    def testseq_slice(self):
        n = 1000
        t = range_tree(0, n)
        for a in range(n):
            u = t[:a]
            self.assertTrue(verify_len(u, a))
            self.assertTrue(equal_tree(u, range_tree(0, a)))
            u = t[a:]
            self.assertTrue(equal_tree(u, range_tree(a, n)))
        t = None

    # test a+b
    def testseq_sliceconcat(self):
        n = 2000
        e = avl.new()
        t = range_tree(0, n)
        self.assertTrue(verify_empty(e+e))
        self.assertTrue(equal_tree(t+e, t))
        self.assertTrue(equal_tree(e+t, t))
        for a in gen_ints_perm(50, n):
            u = t[:a] + t[a:]
            self.assertTrue(u.verify() == 1)
            self.assertTrue(equal_tree(t, u))
            u = None

    def testseq_concatinplace(self):
        self.assertRaises(TypeError, avl.new().concat, [])
        n = 2000
        e = avl.new()
        t = range_tree(0, n)
        u = t[:]
        u.concat(e)
        self.assertTrue(equal_tree(t, u))
        e.concat(u)
        self.assertTrue(equal_tree(e, u))
        u += avl.new()
        self.assertTrue(equal_tree(t, u))
        e.clear()
        e += t
        self.assertTrue(equal_tree(e, t))
        e.clear()
        for a in gen_ints_perm(100, n):
            u = t[:a]
            u += t[a:]
            self.assertTrue(u.verify() == 1)
            self.assertTrue(equal_tree(t, u))
            u = None


# ----
# Test serialization
# ----
def revlowercmp(s1, s2):
    return cmp(string.lower(s2), string.lower(s1))


class Test_avl_pickling(unittest.TestCase):

    def setUp(self):
        self.list1 = []
        self.list2 = []

    def testfrom_iter_basic(self):
        for n in [0, 1, 10, 100, 1000, 10000]:
            a = list(range(n))
            self.assertRaises(AttributeError, avl.from_iter, a, len(a)+1)
            self.assertRaises(StopIteration, avl.from_iter, iter(a), len(a)+1)
        for n in [0, 1, 10, 100, 1000, 10000, 100000]:
            a = list(range(n))
            t = avl.from_iter(iter(a), len(a))
            self.assertTrue(verify_len(t, n))
            for j, k in gen_iter(iter(a), iter(t)):
                self.assertTrue(j == k)

    def testfrom_iter_compare(self):
        a = self.list1 + [string.lower(s) for s in self.list1]
        t = avl.new(a, unique=0, compare=revlowercmp)
        u = avl.from_iter(iter(t), len(t), compare=revlowercmp)
        self.assertTrue(u.verify() == 1)

    def testdump_basic(self):
        t = random_int_tree(0, 3000, 7000)
        for proto in [0, 2]:
            f = BytesIO()
            p = pickle.Pickler(f, proto)
            avl.dump(t, p)
            f.seek(0)
            p = pickle.Unpickler(f)
            a = avl.load(p)
            self.assertTrue(a.verify() == 1)
            self.assertTrue(equal_tree(a, t))
            f.close()

    def testdump_compare(self):
        t = avl.new(self.list1, compare=revlowercmp)
        for proto in [0, 2]:
            f = BytesIO()
            p = pickle.Pickler(f, proto)
            t.dump(p)
            f.seek(0)
            p = pickle.Unpickler(f)
            u = avl.load(p)
            self.assertTrue(u.verify() == 1)
            self.assertTrue(equal_tree(t, u))
            f.close()

    def tearDown(self):
        self.list1 = None


def suite():
    suite = unittest.TestSuite()
    for testcase in [
            Test_avl_ins,
            Test_avl_lookup,
            Test_avl_iter,
            Test_avl_dup,
            Test_avl_del,
            Test_avl_sequence,
            Test_avl_pickling]:
        suite.addTest(unittest.makeSuite(testcase, 'test'))
    return suite


if __name__ == '__main__':
    unittest.main()



# python 2/3
from __future__ import division, print_function, absolute_import
import sys
import unittest
import os
import glob

from . import common

from pyrocko import ExternalProgramMissing

from pyrocko import util
from pyrocko import example
from pyrocko import pile

from pyrocko.gui import snuffler
from pyrocko.dataset import topo


op = os.path

test_dir = op.dirname(op.abspath(__file__))


skip_examples = [
    'examples/trace_restitution_dseed.py'
]


def tutorial_run_dir():
    return op.join(test_dir, 'example_run_dir')


def noop(*args, **kwargs):
    pass


class ExamplesTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        from matplotlib import pyplot as plt

        cls.cwd = os.getcwd()
        cls.run_dir = tutorial_run_dir()
        util.ensuredir(cls.run_dir)
        cls.dn = open(os.devnull, 'w')
        sys.stdout = cls.dn
        os.chdir(cls.run_dir)

        cls._mpl_show_orig = plt.show
        plt.show = noop

        cls._snuffle_orig = snuffler.snuffle
        snuffler.snuffle = noop

        cls._show_progress_force_off_orig = pile.show_progress_force_off
        pile.show_progress_force_off = True

    @classmethod
    def tearDownClass(cls):
        from matplotlib import pyplot as plt

        cls.dn.close()
        sys.stdout = sys.__stdout__
        os.chdir(cls.cwd)

        snuffler.snuffle = cls._snuffle_orig
        plt.show = cls._mpl_show_orig
        pile.show_progress_force_off = cls._show_progress_force_off_orig


example_files = [fn for fn in glob.glob(op.join(test_dir, 'examples', '*.py'))
                 if fn not in skip_examples]


def _make_function(test_name, fn):
    def f(self):
        imp = imp2 = None
        try:
            import imp

        except ImportError:
            import importlib.machinery as imp2

        try:
            if imp:
                imp.load_source(test_name, fn)
            else:
                imp2.SourceFileLoader(test_dir, fn)

        except example.util.DownloadError:
            raise unittest.SkipTest('could not download required data file')

        except ExternalProgramMissing as e:
            raise unittest.SkipTest(str(e))

        except ImportError as e:
            raise unittest.SkipTest(str(e))

        except topo.AuthenticationRequired as e:
            raise unittest.SkipTest('cannot download topo data (no auth credentials)')

        except Exception as e:
            raise e

    f.__name__ = 'test_example_' + test_name

    return f


for fn in sorted(example_files):
    test_name = op.splitext(op.split(fn)[-1])[0]
    setattr(
        ExamplesTestCase,
        'test_example_' + test_name,
        _make_function(test_name, fn))


if __name__ == '__main__':
    util.setup_logging('test_examples', 'warning')
    common.matplotlib_use_agg()
    unittest.main()



from __future__ import division, print_function, absolute_import
import unittest
import numpy as num
from pyrocko import util, config
from pyrocko.dataset import topo


def have_srtm_credentials():
    if config.config().earthdata_credentials is None:
        return False
    return True


class TopoTestCase(unittest.TestCase):

    @unittest.skipUnless(
        have_srtm_credentials(),
        'No Earthdata credentials in config.')
    def test_srtm(self):
        srtm = topo.srtmgl3
        tiles = list(srtm.available_tilenames())

        tilenum = num.random.randint(0, len(tiles)-1)
        srtm.download_tile(tiles[tilenum])
        srtm.get_tile(0, 0)

    @unittest.skip('etopo not downloaded.')
    def test_etopo(self):
        topo.etopo1.make_tiles()

    def test_tile(self):
        tile1 = topo.tile.Tile(
            0., 0.,
            1., 1.,
            num.ones((100, 100)))
        tile2 = topo.tile.Tile(
            0., 0.,
            1., 1.,
            num.ones((100, 100)))

        topo.tile.combine([tile1, tile2])


if __name__ == '__main__':
    util.setup_logging('test_topo', 'debug')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import

import logging
import io
import struct
import time
import numpy as num

from os import path

from pyrocko import config, orthodrome

logger = logging.getLogger('pyrocko.dataset.gshhg')
config = config.config()

km = 1e3
micro_deg = 1e-6


class Polygon(object):
    '''Representation of a GSHHG polygon. '''
    RIVER_NOT_SET = 0

    LEVELS = ['LAND', 'LAKE', 'ISLAND_IN_LAKE', 'POND_IN_ISLAND_IN_LAKE',
              'ANTARCTIC_ICE_FRONT', 'ANTARCTIC_GROUNDING_LINE']

    SOURCE = ['CIA_WDBII', 'WVS', 'AC']

    def __init__(self, gshhg_file, offset, *attr):
        '''Initialise a GSHHG polygon

        :param gshhg_file: GSHHG binary file
        :type gshhg_file: str
        :param offset: This polygons' offset in binary file
        :type offset: int
        :param attr: Polygon attributes
             ``(pid, npoints, _flag, west, east, south, north,
              area, area_full, container, ancestor)``.
              See :file:`gshhg.h` for details.
        :type attr: tuple
        '''
        (self.pid, self.npoints, self._flag,
         self.west, self.east, self.south, self.north,
         self.area, self.area_full, self.container, self.ancestor) = attr

        self.west *= micro_deg
        self.east *= micro_deg
        self.south *= micro_deg
        self.north *= micro_deg

        self.level_no = (self._flag & 255)
        self.level = self.LEVELS[self.level_no - 1]
        self.version = (self._flag >> 8) & 255

        cross = (self._flag >> 16) & 3
        self.greenwhich_crossed = True if cross == 1 or cross == 3 else False
        self.dateline_crossed = True if cross == 2 or cross == 3 else False

        self.source = self.SOURCE[(self._flag >> 24) & 1]
        if self.level_no >= 5:
            self.source = self.SOURCE[2]

        self.river = (self._flag >> 25) & 1

        scale = 10.**(self._flag >> 26)
        self.area /= scale
        self.area_full /= scale

        self._points = None
        self._file = gshhg_file
        self._offset = offset

    @property
    def points(self):
        '''Points of the polygon as Nx2 as ``[:,[lat,lon]]``

        :rtype: :class:`numpy.ndarray`
        '''
        if self._points is None:
            with open(self._file) as db:
                db.seek(self._offset)
                self._points = num.fromfile(
                    db, dtype='>i4', count=self.npoints*2)\
                    .astype(num.float32)\
                    .reshape(self.npoints, 2)

            self._points = num.fliplr(self._points)
            if self.level_no in (2, 4):
                self._points = self._points[::-1, :]

            self._points *= micro_deg
        return self._points

    @property
    def lats(self):
        return self.points[:, 0]

    @property
    def lons(self):
        return self.points[:, 1]

    def _is_level(self, level):
        if self.level is self.LEVELS[level]:
            return True
        return False

    def is_land(self):
        ''' Check if the polygon is land.

        :rtype: bool
        '''
        return self._is_level(0)

    def is_lake(self):
        ''' Check if the polygon is a lake.

        :rtype: bool
        '''
        return self._is_level(1)

    def is_island_in_lake(self):
        ''' Check if the polygon is an island in a lake.

        :rtype: bool
        '''
        return self._is_level(2)

    def is_pond_in_island_in_lake(self):
        ''' Check if the polygon is pond on an island in a lake.

        :rtype: bool
        '''
        return self._is_level(3)

    def is_antarctic_icefront(self):
        ''' Check if the polygon antarctic icefront.

        :rtype: bool
        '''
        return self._is_level(4)

    def is_antarctic_grounding_line(self):
        ''' Check if the polygon is antarctic grounding line.
        :rtype: bool
        '''
        return self._is_level(5)

    def contains_point(self, point):
        ''' Check if the polygon contains a single `point`

        :param point: (lat, lon) of size 2
        :type point: tuple
        :rtype: bool
        '''
        if self.south <= point[0] <= self.north and\
           self.west <= point[1] <= self.east:
            return orthodrome.contains_point(self.points, point)
        return False

    def contains_points(self, points):
        ''' Check if the polygon contains a `points`

        :param points: Array of of size Nx2
        :type point: :class:`numpy.ndarray`
        :rtype: bool
        '''
        cond = num.all([self.south <= points[:, 0],
                        points[:, 0] <= self.north,
                        self.west <= points[:, 1],
                        points[:, 1] <= self.east],
                       axis=0)
        r = orthodrome.contains_points(self.points, points)
        logger.debug('%s: points inside %d' % (self.level, r.sum()))
        return r
        if num.any(cond):
            pass

        return num.zeros(points.shape[0], dtype=num.bool)

    def get_bounding_box(self):
        return (self.west, self.east, self.south, self.north)

    def __lt__(self, polygon):
        return self.level_no < polygon.level_no

    def __str__(self):
        rstr = '''Polygon id: {p.pid}
-------------------
Points:     {p.npoints}
Level:      {p.level}
Area:       {p.area} km**2
Area Full:  {p.area_full} km**2
Extent:     {p.west} W, {p.east} E, {p.south} S, {p.north} N
Source:     {p.source}
Greenwhich crossed: {p.greenwhich_crossed}
Dateline crossed:   {p.dateline_crossed}
        '''.format(p=self)
        return rstr


class GSHHG(object):
    '''Holding the Global Self-consistent Hierarchical High-resolutions
        Geography Database (GSHHG)

    The class offers methods to select :class:`~pyrocko.gshhg.Polygon` s and
    crop/validate single points and point-clouds.

    .. info:

        If the database is not available it will be downloaded and cached
        automatically

    .. note:

        Cite Wessel, P., and W. H. F. Smith, A Global Self-consistent,
        Hierarchical, High-resolution Shoreline Database,
        J. Geophys. Res., 101, #B4, pp. 8741-8743, 1996.
    '''

    gshhg_url = 'http://www.soest.hawaii.edu/pwessel/gshhg/gshhg-bin-2.3.7.zip'
    _header_struct = struct.Struct('>IIIiiiiIIii')

    def __init__(self, gshhg_file):
        ''' Initialise the database from GSHHG binary.

        :param gshhg_file: Path to file
        :type gshhg_file: str
        :
        '''
        t0 = time.time()
        self._file = gshhg_file

        self.polygons = []
        self._read_database()
        logger.debug('Initialised GSHHG database from %s in [%.4f s]'
                     % (gshhg_file, time.time()-t0))

    def _read_database(self):
        with open(self._file, mode='rb') as db:
            while db:
                buf = db.read(self._header_struct.size)
                if not buf:
                    break
                header = self._header_struct.unpack_from(buf)
                p = Polygon(
                    self._file,
                    db.tell(),
                    *header)
                self.polygons.append(p)

                offset = 8 * header[1]
                db.seek(offset, io.SEEK_CUR)

    @classmethod
    def _get_database(cls, filename):
        file = path.join(config.gshhg_dir, filename)
        if not path.exists(file):
            from pyrocko import util
            import zipfile

            archive_path = path.join(config.gshhg_dir,
                                     path.basename(cls.gshhg_url))
            util.download_file(cls.gshhg_url, archive_path)
            if not zipfile.is_zipfile(archive_path):
                raise util.DownloadError('GSHHG file is corrupted!')
            logger.info('Unzipping GSHHG database...')
            zipf = zipfile.ZipFile(archive_path)
            zipf.extractall(config.gshhg_dir)
        else:
            logger.debug('Using cached %s' % filename)
        return file

    def get_polygons_at(self, lat, lon):
        '''Get all polygons that intersect with a point.

        :param lat: Latitude in [deg]
        :type lat: float
        :param lat: Longitude in [deg]
        :type lat: float
        :returns: List of :class:`~pyrocko.gshhg.Polygon`
        :rtype: list
        '''
        rp = []
        for p in self.polygons:
            if (p.west < lon and p.east > lon) and\
               (p.south < lat and p.north > lat):
                rp.append(p)
        return rp

    def get_polygons_within(self, west, east, south, north):
        '''Get all polygons that intersect with a bounding box.

        :param west: Western boundary in decimal degree
        :type west: float
        :param east: Eastern boundary in decimal degree
        :type east: float
        :param north: Northern boundary in decimal degree
        :type north: float
        :param south: Southern boundary in decimal degree
        :type south: float
        :returns: List of :class:`~pyrocko.gshhg.Polygon`
        :rtype: list
        '''
        rp = []
        for p in self.polygons:
            if ((p.west > west and p.east < east) or
               (p.west < west and p.east > west) or
               (p.west < east and p.east > east) or
               (p.west > west and p.east < east)) and\
               ((p.south > south and p.north < north) or
               (p.south < south and p.north > south) or
               (p.south < north and p.north > north) or
               (p.north > north and p.south < south)):
                rp.append(p)
        return rp

    def is_point_on_land(self, lat, lon):
        '''Check whether a point is on land. Consquently lakes are excluded.

        :param lat: Latitude in [deg]
        :type lat: float
        :param lon: Latitude in [deg]
        :type lon: float
        :rtype: bool
        '''
        relevant_polygons = self.get_polygons_at(lat, lon)
        relevant_polygons.sort()

        land = False
        for p in relevant_polygons:
            if (p.is_land() or p.is_antarctic_grounding_line() or
               p.is_island_in_lake()):
                land = True
            elif (p.is_lake() or p.is_antarctic_icefront() or
                  p.is_pond_in_island_in_lake()):
                land = False
        return land

    def get_land_mask(self, points):
        '''Get a landmask respecting lakes, and ponds in island in lake
            as water

        :param points: List of lat, lon pairs
        :type points: :class:`numpy.ndarray` of shape Nx2
        :return: Boolean land mask
        :rtype: :class:`numpy.ndarray` of shape N
        '''

        lats = points[:, 0]
        lons = points[:, 1]
        west, east, south, north = (lons.min(), lons.max(),
                                    lats.min(), lats.max())

        relevant_polygons = self.get_polygons_within(west, east, south, north)
        relevant_polygons.sort()

        mask = num.zeros(points.shape[0], dtype=num.bool)
        for p in relevant_polygons:
            if (p.is_land() or p.is_antarctic_grounding_line() or
               p.is_island_in_lake()):
                mask += p.contains_points(points)
            elif p.is_lake() or p.is_pond_in_island_in_lake():
                water = p.contains_points(points)
                num.logical_xor(mask, water, out=mask)
        return mask

    @classmethod
    def full(cls):
        ''' Return the full-resolution GSHHG database'''
        return cls(cls._get_database('gshhs_f.b'))

    @classmethod
    def high(cls):
        ''' Return the high-resolution GSHHG database'''
        return cls(cls._get_database('gshhs_h.b'))

    @classmethod
    def intermediate(cls):
        ''' Return the intermediate-resolution GSHHG database'''
        return cls(cls._get_database('gshhs_i.b'))

    @classmethod
    def low(cls):
        ''' Return the low-resolution GSHHG database'''
        return cls(cls._get_database('gshhs_l.b'))

    @classmethod
    def crude(cls):
        ''' Return the crude-resolution GSHHG database'''
        return cls(cls._get_database('gshhs_c.b'))



import logging
import os.path as op
import numpy as num

from pyrocko import gf
from pyrocko.guts import Float

from .base import TargetGenerator, NoiseGenerator
from .station import RandomStationGenerator, StationGenerator

DEFAULT_STORE_ID = 'ak135_static'

logger = logging.getLogger('pyrocko.scenario.targets.gnss_campaign')
guts_prefix = 'pf.scenario'


class GPSNoiseGenerator(NoiseGenerator):
    measurement_duarion_days = Float.T(
        default=2.,
        help='Measurement duration in days')

    def add_noise(self, campaign):
        # https://www.nat-hazards-earth-syst-sci.net/15/875/2015/nhess-15-875-2015.pdf
        waterlevel = 1. - (.99 + .0015 * self.measurement_duarion_days)  # noqa
        logger.warning('GNSSNoiseGenerator is a work-in-progress!')

        for ista, sta in enumerate(campaign.stations):
            pass
            # rstate = self.get_rstate(ista)

            # sta.north.sigma = 8e-3
            # sta.east.sigma = 8e-3

            # sta.north.shift += rstate.normal(0., sta.north.sigma)
            # sta.east.shift += rstate.normal(0., sta.east.sigma)


class GNSSCampaignGenerator(TargetGenerator):
    station_generator = StationGenerator.T(
        default=RandomStationGenerator(
            network_name='GN',
            channels=None),
        help='The StationGenerator for creating the stations.')

    noise_generator = NoiseGenerator.T(
        default=GPSNoiseGenerator.D(),
        optional=True,
        help='Add Synthetic noise to the GNSS displacements.')

    store_id = gf.StringID.T(
        default=DEFAULT_STORE_ID,
        help='The GF store to use for forward-calculations.')

    def get_stations(self):
        return self.station_generator.get_stations()

    def get_targets(self):
        stations = self.get_stations()
        lats = num.array([s.lat for s in stations])
        lons = num.array([s.lon for s in stations])

        target = gf.GNSSCampaignTarget(
            lats=lats,
            lons=lons,
            store_id=self.store_id)

        return [target]

    def get_gnss_campaigns(self, engine, sources, tmin=None, tmax=None):
        try:
            resp = engine.process(
                sources,
                self.get_targets(),
                nthreads=0)
        except gf.meta.OutOfBounds:
            logger.warning('Could not calculate GNSS displacements'
                           ' - the GF store\'s extend is too small!')
            return []

        campaigns = [r.campaign for r in resp.static_results()]

        stacked_campaign = campaigns[0]
        stacked_campaign.name = 'Scenario Campaign'
        for camp in campaigns[1:]:
            for ista, sta in enumerate(camp.stations):
                stacked_campaign.stations[ista].north.shift += sta.north.shift
                stacked_campaign.stations[ista].east.shift += sta.east.shift
                stacked_campaign.stations[ista].up.shift += sta.up.shift

        for ista, sta in enumerate(stacked_campaign.stations):
            sta.code = 'SY%02d' % (ista + 1)

        if self.noise_generator is not None:
            self.noise_generator.add_noise(stacked_campaign)

        return [stacked_campaign]

    def dump_data(self, engine, sources, path,
                  tmin=None, tmax=None, overwrite=False):
        path_gnss = op.join(path, 'gnss')
        gf.store.remake_dir(path_gnss, force=overwrite)

        campaigns = self.get_gnss_campaigns(
            engine, sources, tmin, tmax)

        fn = op.join(path_gnss,
                     'campaign-%s.yml' % self.station_generator.network_name)

        with open(fn, 'w') as f:
            for camp in campaigns:
                camp.dump(stream=f)

        return [fn]

    def add_map_artists(self, engine, sources, automap):
        automap.add_gnss_campaign(self.get_gnss_campaigns(engine, sources)[0])



from __future__ import division, print_function, absolute_import
import unittest
try:
    from . import common
except Exception:
    import common
import numpy as num
import tempfile
import os

from pyrocko import util, model
from pyrocko.pile import make_pile
from pyrocko import config, trace

if common.have_gui():  # noqa
    from pyrocko.gui.qt_compat import qc, qw, use_pyqt5
    if use_pyqt5:
        from PyQt5.QtTest import QTest
        Qt = qc.Qt
    else:
        from PyQt4.QtTest import QTest
        Qt = qc.Qt

    from pyrocko.gui.snuffler_app import Snuffler, SnufflerWindow
    from pyrocko.gui import pile_viewer as pyrocko_pile_viewer
    from pyrocko.gui import util as gui_util
    from pyrocko.gui import snuffling

    class TestSnuffling(snuffling.Snuffling):

        def setup(self):
            self.set_name('TestSnuffling')

        def call(self):
            figframe = self.figure_frame()
            ax = figframe.gca()
            ax.plot([0, 1], [0, 1])
            figframe.draw()

            self.enable_pile_changed_notifications()

            self.pixmap_frame()
            self.web_frame()

            self.get_pile()

    no_gui = False
else:
    no_gui = True


@common.require_gui
class GUITest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        '''
        Create a reusable snuffler instance for all tests cases.
        '''
        super(GUITest, cls).setUpClass()
        if no_gui:  # nosetests runs this even when class is has @skip
            return

        cls.snuffler = Snuffler()  # noqa
        fpath = common.test_data_file('test2.mseed')
        p = make_pile(fpath, show_progress=False)
        cls.win = SnufflerWindow(pile=p)
        cls.pile_viewer = cls.win.pile_viewer
        cls.viewer = cls.win.pile_viewer.viewer
        pv = cls.pile_viewer
        cls.main_control_defaults = dict(
            highpass_control=pv.highpass_control.get_value(),
            lowpass_control=pv.lowpass_control.get_value(),
            gain_control=pv.gain_control.get_value(),
            rot_control=pv.rot_control.get_value())

    @classmethod
    def tearDownClass(cls):
        '''
        Quit snuffler.
        '''
        if no_gui:  # nosetests runs this even when class is has @skip
            return

        QTest.keyPress(cls.pile_viewer, 'q')

    def setUp(self):
        '''
        reset GUI
        '''
        for k, v in self.main_control_defaults.items():
            getattr(self.pile_viewer, k).set_value(v)

        self.initial_trange = self.viewer.get_time_range()
        self.viewer.set_tracks_range(
            [0, self.viewer.ntracks_shown_max])
        self.tempfiles = []

    def tearDown(self):
        self.clear_all_markers()
        for tempfn in self.tempfiles:
            os.remove(tempfn)

        self.viewer.set_time_range(*self.initial_trange)

    def get_tempfile(self):
        tempfn = tempfile.mkstemp()[1]
        self.tempfiles.append(tempfn)
        return tempfn

    def write_to_input_line(self, text):
        '''emulate writing to inputline and press return'''
        pv = self.pile_viewer
        il = pv.inputline
        QTest.keyPress(pv, ':')
        QTest.keyClicks(il, text)
        QTest.keyPress(il, Qt.Key_Return)

    def clear_all_markers(self):
        pv = self.pile_viewer
        QTest.keyPress(pv, 'A', Qt.ShiftModifier, 10)
        QTest.keyPress(pv, Qt.Key_Backspace)
        self.assertEqual(len(pv.viewer.get_markers()), 0)

    def trigger_menu_item(self, qmenu, action_text, dialog=False):
        ''' trigger a QMenu QAction with action_text. '''

        for iaction, action in enumerate(qmenu.actions()):
            if action.text() == action_text:

                if dialog:
                    def closeDialog():
                        dlg = self.snuffler.activeModalWidget()
                        QTest.keyClick(dlg, Qt.Key_Escape)

                    qc.QTimer.singleShot(150, closeDialog)

                action.trigger()
                break

    def get_slider_position(self, slider):
        style = slider.style()
        opt = qw.QStyleOptionSlider()
        return style.subControlRect(
            qw.QStyle.CC_Slider, opt, qw.QStyle.SC_SliderHandle)

    def drag_slider(self, slider):
        ''' Click *slider*, drag from one side to the other, release mouse
        button repeat to restore inital state'''
        position = self.get_slider_position(slider)
        QTest.mouseMove(slider, pos=position.topLeft())
        QTest.mousePress(slider, Qt.LeftButton)
        QTest.mouseMove(slider, pos=position.bottomRight())
        QTest.mouseRelease(slider, Qt.LeftButton)
        QTest.mousePress(slider, Qt.LeftButton)
        QTest.mouseMove(slider, pos=position.topLeft())
        QTest.mouseRelease(slider, Qt.LeftButton)

    def add_one_pick(self):
        '''Add a single pick to pile_viewer'''
        pv = self.pile_viewer
        QTest.mouseDClick(pv.viewer, Qt.LeftButton)
        position_tl = pv.pos()
        geom = pv.frameGeometry()
        QTest.mouseMove(pv.viewer, pos=position_tl)
        QTest.mouseMove(pv.viewer, pos=(qc.QPoint(
            position_tl.x()+geom.x()/2., position_tl.y()+geom.y()/2.)))

        # This should be done also by mouseDClick().
        QTest.mouseRelease(pv.viewer, Qt.LeftButton)
        QTest.mouseClick(pv.viewer, Qt.LeftButton)

    def test_main_control_sliders(self):
        self.drag_slider(self.pile_viewer.highpass_control.slider)
        self.drag_slider(self.pile_viewer.lowpass_control.slider)
        self.drag_slider(self.pile_viewer.gain_control.slider)
        self.drag_slider(self.pile_viewer.rot_control.slider)

    def test_inputline(self):
        initrange = self.viewer.shown_tracks_range

        self.write_to_input_line('hide W.X.Y.Z')
        self.write_to_input_line('unhide W.X.Y.Z')
        self.pile_viewer.update()

        self.write_to_input_line('hide *')
        self.pile_viewer.update()

        assert(self.viewer.shown_tracks_range == (0, 1))
        self.write_to_input_line('unhide')

        assert(self.viewer.shown_tracks_range == initrange)

        self.write_to_input_line('markers')
        self.write_to_input_line('markers 4')
        self.write_to_input_line('markers all')

        # should error
        self.write_to_input_line('scaling 1000.')
        self.write_to_input_line('scaling -1000. 1000.')

        gotos = ['2015-01-01 00:00:00',
                 '2015-01-01 00:00',
                 '2015-01-01 00',
                 '2015-01-01',
                 '2015-01',
                 '2015']

        for gt in gotos:
            self.write_to_input_line('goto %s' % gt)

        # test some false input
        self.write_to_input_line('asdf')
        QTest.keyPress(self.pile_viewer.inputline, Qt.Key_Escape)

    def test_drawing_optimization(self):
        n = 505
        lats = num.random.uniform(-90., 90., n)
        lons = num.random.uniform(-180., 180., n)
        events = []
        for i, (lat, lon) in enumerate(zip(lats, lons)):
            events.append(
                model.Event(time=i, lat=lat, lon=lon, name='XXXX%s' % i))

        self.viewer.add_event(events[-1])
        assert len(self.viewer.markers) == 1
        self.viewer.add_events(events)
        assert len(self.viewer.markers) == n + 1

        self.viewer.set_time_range(-500., 5000)
        self.viewer.set_time_range(0., None)
        self.viewer.set_time_range(None, 0.)

    def test_follow(self):
        self.viewer.follow(10.)
        self.viewer.unfollow()

    def test_save_image(self):
        tempfn_svg = self.get_tempfile() + '.svg'
        self.viewer.savesvg(fn=tempfn_svg)

        tempfn_png = self.get_tempfile() + '.png'
        self.viewer.savesvg(fn=tempfn_png)

    def test_read_events(self):
        event = model.Event()
        tempfn = self.get_tempfile()
        model.event.dump_events([event], tempfn)
        self.viewer.read_events(tempfn)

    def test_add_remove_stations(self):
        n = 10
        lats = num.random.uniform(-90., 90., n)
        lons = num.random.uniform(-180., 180., n)
        stations = [
            model.station.Station(network=str(i), station=str(i),
                                  lat=lat, lon=lon) for i, (lat, lon) in
            enumerate(zip(lats, lons))
        ]
        tempfn = self.get_tempfile()
        model.station.dump_stations(stations, tempfn)
        self.viewer.open_stations(fns=[tempfn])
        last = stations[-1]
        self.assertTrue(self.viewer.has_station(last))
        self.viewer.get_station((last.network, last.station))

    def test_markers(self):
        self.add_one_pick()
        pv = self.pile_viewer
        self.assertEqual(pv.viewer.get_active_event(), None)

        conf = config.config('snuffler')

        # test kinds and phases
        kinds = range(5)
        fkey_map = pyrocko_pile_viewer.fkey_map

        for k in kinds:
            for fkey, fkey_int in fkey_map.items():
                fkey_int += 1
                QTest.keyPress(pv, fkey)
                QTest.keyPress(pv, str(k))

                if fkey_int != 10:
                    want = conf.phase_key_mapping.get(
                        "F%s" % fkey_int, 'Undefined')
                else:
                    want = None
                m = pv.viewer.get_markers()[0]
                self.assertEqual(m.kind, k)
                if want:
                    self.assertEqual(m.get_phasename(), want)

    def test_load_waveforms(self):
        self.viewer.load('data', regex='\w*.mseed')
        self.assertFalse(self.viewer.get_pile().is_empty())

    def test_add_traces(self):
        trs = []
        for i in range(3):
            trs.append(
                trace.Trace(network=str(i), tmin=num.random.uniform(1),
                            ydata=num.random.random(100),
                            deltat=num.random.random())
            )
        self.viewer.add_traces(trs)

    def test_event_marker(self):
        pv = self.pile_viewer
        self.add_one_pick()

        # select all markers
        QTest.keyPress(pv, 'a', Qt.ShiftModifier, 100)

        # convert to EventMarker
        QTest.keyPress(pv, 'e')

        QTest.keyPress(pv, 'd')

        for m in pv.viewer.get_markers():
            self.assertTrue(isinstance(m, gui_util.EventMarker))

    def test_load_save_markers(self):
        nmarkers = 505
        times = num.arange(nmarkers)
        markers = [gui_util.Marker(tmin=t, tmax=t,
                                   nslc_ids=[('*', '*', '*', '*'), ])
                   for t in times]

        tempfn = self.get_tempfile()
        tempfn_selected = self.get_tempfile()

        self.viewer.add_markers(markers)
        self.viewer.write_selected_markers(
            fn=tempfn_selected)
        self.viewer.write_markers(fn=tempfn)

        self.viewer.read_markers(fn=tempfn_selected)
        self.viewer.read_markers(fn=tempfn)

        for k in 'pnPN':
            QTest.keyPress(self.pile_viewer, k)

        self.viewer.go_to_time(-20., 20)
        self.pile_viewer.update()
        self.viewer.update()
        assert(len(self.viewer.markers) != 0)
        assert(len(self.viewer.markers) == nmarkers * 2)
        len_before = len(self.viewer.markers)
        self.viewer.remove_marker(
            self.viewer.markers[0])
        assert(len(self.viewer.markers) == len_before-1)
        self.viewer.remove_markers(self.viewer.markers)
        assert(len(self.viewer.markers) == 0)

    def test_actions(self):
        # Click through many menu option combinations that do not require
        # further interaction. Activate options in pairs of two.

        pv = self.pile_viewer
        tinit = pv.viewer.tmin
        tinitlen = pv.viewer.tmax - pv.viewer.tmin

        non_dialog_actions = [
            'Indivdual Scale',
            'Common Scale',
            'Common Scale per Station',
            'Common Scale per Component',
            'Scaling based on Minimum and Maximum',
            'Scaling based on Mean +- 2 x Std. Deviation',
            'Scaling based on Mean +- 4 x Std. Deviation',
            'Sort by Names',
            'Sort by Distance',
            'Sort by Azimuth',
            'Sort by Distance in 12 Azimuthal Blocks',
            'Sort by Backazimuth',
            '3D distances',
            'Subsort by Network, Station, Location, Channel',
            'Subsort by Network, Station, Channel, Location',
            'Subsort by Station, Network, Channel, Location',
            'Subsort by Location, Network, Station, Channel',
            'Subsort by Channel, Network, Station, Location',
            'Subsort by Network, Station, Channel (Grouped by Location)',
            'Subsort by Station, Network, Channel (Grouped by Location)',
        ]

        dialog_actions = [
            'Open waveform files...',
            'Open waveform directory...',
            'Open station files...',
            'Save markers...',
            'Save selected markers...',
            'Open marker file...',
            'Open event file...',
            'Save as SVG|PNG',
            ]

        options = [
            'Antialiasing',
            'Liberal Fetch Optimization',
            'Clip Traces',
            'Show Boxes',
            'Color Traces',
            'Show Scale Ranges',
            'Show Scale Axes',
            'Show Zero Lines',
            'Fix Scale Ranges',
            'Allow Downsampling',
            'Allow Degapping',
            'FFT Filtering',
            'Bandpass is Lowpass + Highpass',
            'Watch Files',
        ]

        # create an event marker and activate it
        self.add_one_pick()

        keys = list('mAhefrRh+-fgc?')
        keys.extend([Qt.Key_PageUp, Qt.Key_PageDown])

        def fire_key(x):
            QTest.keyPress(self.pile_viewer, key)

        for key in keys:
            QTest.qWait(100)
            fire_key(key)

        event = model.Event()
        markers = pv.viewer.get_markers()
        self.assertEqual(len(markers), 1)
        markers[0]._event = event

        pv.viewer.set_active_event(event)
        pv.viewer.set_event_marker_as_origin()

        right_click_menu = self.viewer.menu

        for action_text in dialog_actions:
            self.trigger_menu_item(right_click_menu, action_text, dialog=True)

        for action_text in non_dialog_actions:
            for oa in options:
                for ob in options:
                    self.trigger_menu_item(right_click_menu, action_text)
                    self.trigger_menu_item(right_click_menu, oa)
                    self.trigger_menu_item(right_click_menu, ob)

                options.remove(oa)

        self.viewer.go_to_event_by_name(event.name)
        self.viewer.go_to_time(tinit, tinitlen)

    def test_frames(self):
        frame_snuffling = TestSnuffling()

        self.viewer.add_snuffling(frame_snuffling)
        frame_snuffling.call()

        # close three opened frames
        QTest.keyPress(self.pile_viewer, 'd')
        QTest.keyPress(self.pile_viewer, 'd')
        QTest.keyPress(self.pile_viewer, 'd')


if __name__ == '__main__':
    util.setup_logging('test_gui', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range, zip

import random
import math
import unittest
import logging
from tempfile import mkdtemp
import numpy as num
import copy
import os
import shutil

from pyrocko import util, trace, gf, cake  # noqa
from pyrocko.fomosto import qseis
from pyrocko.fomosto import qseis2d

logger = logging.getLogger('pyrocko.test.test_qseis_qseis2d')

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1e3
slowness_window = (0.0, 0.0, 0.4, 0.5)


@unittest.skipUnless(
    qseis2d.have_backend(), 'backend qseis2d not available')
@unittest.skipUnless(
    qseis.have_backend(), 'backend qseis not available')
class GFQSeis2dTestCase(unittest.TestCase):

    tempdirs = []

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def test_pyrocko_gf_vs_qseis2d(self):

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))

        receiver_mod = cake.LayeredModel.from_scanlines(
                                    cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
'''.lstrip()))

        q2_store_dir = mkdtemp(prefix='gfstore')
        q_store_dir = mkdtemp(prefix='gfstore')
        self.tempdirs.append(q2_store_dir)
        self.tempdirs.append(q_store_dir)

        # qseis2d
        q2conf = qseis2d.QSeis2dConfig()
        q2conf.gf_directory = os.path.join(q2_store_dir, 'qseisS_green')

        qss = q2conf.qseis_s_config
        qss.receiver_basement_depth = 35.
        qss.calc_slowness_window = 0
        qss.slowness_window = slowness_window

        q2conf.time_region = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        q2conf.cut = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qss.sw_flat_earth_transform = 0

        config_q2 = gf.meta.ConfigTypeA(
            id='qseis2d_test_q2',
            ncomponents=10,
            sample_rate=0.25,
            receiver_depth=0. * km,
            source_depth_min=10. * km,
            source_depth_max=10. * km,
            source_depth_delta=1. * km,
            distance_min=5528. * km,
            distance_max=5532. * km,
            distance_delta=1. * km,
            modelling_code_id='qseis2d',
            earthmodel_1d=mod,
            earthmodel_receiver_1d=receiver_mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='2.5'),
            ])

        config_q2.validate()
        gf.store.Store.create_editables(
            q2_store_dir, config=config_q2, extra={'qseis2d': q2conf})

        store = gf.store.Store(q2_store_dir, 'r')
        store.make_ttt()
        store.close()

        # build store
        try:
            qseis2d.build(q2_store_dir, nworkers=1)
        except qseis2d.QSeis2dError as e:
            if str(e).find('could not start qseis2d') != -1:
                logger.warn('qseis2d not installed; '
                            'skipping test_pyrocko_qseis_vs_qseis2d')
                logger.warn(e)
                return
            else:
                raise

        # qseis
        config_q = copy.deepcopy(config_q2)
        config_q.id = 'qseis2d_test_q'
        config_q.modelling_code_id = 'qseis.2006a'

        qconf = qseis.QSeisConfig()
        qconf.qseis_version = '2006a'

        qconf.slowness_window = slowness_window

        qconf.time_region = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qconf.cut = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qconf.sw_flat_earth_transform = 0

        config_q.validate()
        gf.store.Store.create_editables(
            q_store_dir, config=config_q, extra={'qseis': qconf})

        store = gf.store.Store(q_store_dir, 'r')
        store.make_ttt()
        store.close()

        # build store
        try:
            qseis.build(q_store_dir, nworkers=1)
        except qseis.QSeisError as e:
            if str(e).find('could not start qseis') != -1:
                logger.warn('qseis not installed; '
                            'skipping test_pyrocko_qseis_vs_qseis2d')
                logger.warn(e)
                return
            else:
                raise

        # forward model
        source = gf.MTSource(
            lat=0.,
            lon=0.,
            depth=10. * km)

        source.m6 = tuple(random.random() * 2. - 1. for x in range(6))

        azi = 0.    # QSeis2d only takes one receiver without azimuth variable
        dist = 5530. * km

        dnorth = dist * math.cos(azi * d2r)
        deast = dist * math.sin(azi * d2r)

        targets_q = []
        targets_q2 = []
        for cha in 'rtz':
            target_q2 = gf.Target(
                quantity='displacement',
                codes=('', '0000', 'Q2', cha),
                north_shift=dnorth,
                east_shift=deast,
                store_id='qseis2d_test_q2')

            dist = source.distance_to(target_q2)
            azi, bazi = source.azibazi_to(target_q2)

            if cha == 'r':
                target_q2.azimuth = bazi + 180.
                target_q2.dip = 0.
            elif cha == 't':
                target_q2.azimuth = bazi - 90.
                target_q2.dip = 0.
            elif cha == 'z':
                target_q2.azimuth = 0.
                target_q2.dip = 90.

            target_q = copy.deepcopy(target_q2)
            target_q.store_id = 'qseis2d_test_q'
            target_q.codes = ('', '0000', 'Q', cha)
            targets_q2.append(target_q2)
            targets_q.append(target_q)

        targets = targets_q + targets_q2
        engine = gf.LocalEngine(store_dirs=[q2_store_dir, q_store_dir])
        response = engine.process(source, targets)

        qtrcs = []
        q2trcs = []
        for s, target, trc in response.iter_results():
            if target.codes[2] == 'Q':
                qtrcs.append(trc)
            else:
                q2trcs.append(trc)

        for q, q2 in zip(qtrcs, q2trcs):
            num.testing.assert_allclose(q.ydata, q2.ydata, atol=4e-24)

#        trace.snuffle(qtrcs + q2trcs)


if __name__ == '__main__':
    util.setup_logging('test_qseis_qseis2d', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import

from builtins import str
from builtins import range
import os
import unittest
import numpy as num
import time
import tempfile
import random
from random import choice as rc
from os.path import join as pjoin
import shutil

from pyrocko import io, guts
from pyrocko.io import FileLoadError
from pyrocko.io import mseed, trace, util, suds, quakeml

from . import common

abc = 'abcdefghijklmnopqrstuvwxyz'


def rn(n):
    return ''.join([random.choice(abc) for i in range(n)])


class IOTestCase(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='pyrocko')

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def testWriteRead(self):
        now = time.time()
        n = 10
        deltat = 0.1

        networks = [rn(2) for i in range(5)]

        traces1 = [
            trace.Trace(
                rc(networks), rn(4), rn(2), rn(3),
                tmin=now+i*deltat*n*2,
                deltat=deltat,
                ydata=num.arange(n, dtype=num.int32),
                mtime=now)

            for i in range(3)]

        for format in ('mseed', 'sac', 'yaff', 'gse2'):
            fns = io.save(
                traces1,
                pjoin(
                    self.tmpdir,
                    '%(network)s_%(station)s_%(location)s_%(channel)s'),
                format=format)

            for fn in fns:
                assert io.detect_format(fn) == format

            traces2 = []
            for fn in fns:
                traces2.extend(io.load(fn, format='detect'))

            for tr in traces1:
                assert tr in traces2, 'failed for format %s' % format

            for fn in fns:
                os.remove(fn)

    def testWriteText(self):
        networks = [rn(2) for i in range(5)]
        deltat = 0.1
        tr = trace.Trace(
            rc(networks), rn(4), rn(2), rn(3),
            tmin=time.time()+deltat,
            deltat=deltat,
            ydata=num.arange(100, dtype=num.int32),
            mtime=time.time())
        io.save(
            tr,
            pjoin(
                self.tmpdir,
                '%(network)s_%(station)s_%(location)s_%(channel)s'),
            format='text')

    def testReadEmpty(self):
        tempfn = tempfile.mkstemp()[1]
        try:
            list(mseed.iload(tempfn))
        except FileLoadError as e:
            assert str(e).find('No SEED data detected') != -1

        os.remove(tempfn)

    def testReadSac(self):
        fpath = common.test_data_file('test1.sac')
        tr = io.load(fpath, format='sac')[0]
        assert tr.meta['cmpaz'] == 0.0
        assert tr.meta['cmpinc'] == 0.0

    def testReadSac2(self):
        fpath = common.test_data_file('test2.sac')
        tr = io.load(fpath, format='sac')[0]
        assert tr.location == ''

    def testLongCode(self):
        c = '1234567'
        tr = trace.Trace(c, c, c, c, ydata=num.zeros(10))
        e = None
        try:
            io.save(tr, 'test.mseed')
        except mseed.CodeTooLong as e:
            assert isinstance(e, mseed.CodeTooLong)

    def testMSeedDetect(self):
        fpath = common.test_data_file('test2.mseed')
        io.load(fpath, format='detect')

    def testReadSEGY(self):
        fpath = common.test_data_file('test2.segy')
        i = 0
        for tr in io.load(fpath, format='segy'):
            assert tr.meta['orfield_num'] == 1111
            i += 1

        assert i == 24

    def testReadGSE1(self):
        fpath = common.test_data_file('test1.gse1')
        i = 0
        for tr in io.load(fpath, format='detect'):
            i += 1

        assert i == 19

    def testReadSUDS(self):
        fpath = common.test_data_file('test.suds')
        i = 0
        for tr in io.load(fpath, format='detect'):
            i += 1

        assert i == 251

        stations = suds.load_stations(fpath)

        assert len(stations) == 91

    def testReadCSS(self):
        wfpath = common.test_data_file('test_css1.w')  # noqa
        fpath = common.test_data_file('test_css.wfdisc')
        i = 0
        for tr in io.load(fpath, format='css'):
            i += 1

        assert i == 1

    def testReadSeisan(self):
        fpath = common.test_data_file('test.seisan_waveform')
        i = 0
        for tr in io.load(fpath, format='seisan'):
            i += 1

        assert i == 39

    def testReadKan(self):
        fpath = common.test_data_file('01.kan')
        i = 0
        for tr in io.load(fpath, format='kan'):
            i += 1

        assert i == 1

    def testReadGcf(self):
        fpath = common.test_data_file('test.gcf')

        i = 0
        for tr in io.load(fpath, format='gcf'):
            i += 1

        assert i == 1

    def testReadQuakeML(self):

        fpath = common.test_data_file('test.quakeml')
        qml = quakeml.QuakeML.load_xml(filename=fpath)
        events = qml.get_pyrocko_events()
        assert len(events) == 1
        e = events[0]
        assert e.lon == -116.9945
        assert e.lat == 33.986
        assert e.depth == 17300
        assert e.time == util.stt("1999-04-02 17:05:10.500")

        fpath = common.test_data_file('example-catalog.xml')
        qml = quakeml.QuakeML.load_xml(filename=fpath)
        events = qml.get_pyrocko_events()
        assert len(events) == 2
        assert events[0].moment_tensor is not None

        s = qml.dump_xml(ns_ignore=True)
        qml2 = quakeml.QuakeML.load_xml(string=s)
        qml3 = guts.load_xml(string=s)
        assert qml2.dump_xml() == qml3.dump_xml()

    def testReadQuakeML2(self):
        for fn in ['usgs.quakeml', 'isc.quakeml']:
            fpath = common.test_data_file(fn)
            qml = quakeml.QuakeML.load_xml(filename=fpath)
            s = qml.dump_xml()

            qml2 = quakeml.QuakeML.load_xml(string=s)
            s2 = qml2.dump_xml()
            assert len(s) == len(s2)

    def testReadStationXML(self):
        from pyrocko.io import stationxml  # noqa

        fpath = common.test_data_file('test1.stationxml')
        sx = guts.load_xml(filename=fpath)
        s = sx.dump_xml(ns_ignore=True)
        sx2 = guts.load_xml(string=s)
        assert sx.dump_xml() == sx2.dump_xml()


if __name__ == "__main__":
    util.setup_logging('test_io', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import sys
import math
import unittest
import numpy as num
from tempfile import mkdtemp
import shutil

from pyrocko import gf, util, guts, cake, moment_tensor as pmt

r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


def numeq(a, b, eps):
    return (num.all(num.asarray(a).shape == num.asarray(b).shape and
            num.abs(num.asarray(a) - num.asarray(b)) < eps))


def default_source(S, **kwargs):
    if S is not gf.CombiSource:
        return S(**kwargs)
    else:
        return S([gf.MTSource(**kwargs)])


class GFSourcesTestCase(unittest.TestCase):
    tempdirs = []

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)
        self._dummy_store = None
        self._dummy_homogeneous_store = None

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    if sys.version_info < (2, 7):
        from contextlib import contextmanager

        @contextmanager
        def assertRaises(self, exc):

            gotit = False
            try:
                yield None
            except exc:
                gotit = True

            assert gotit, 'expected to get a %s exception' % exc

        def assertIsNone(self, value):
            assert value is None, 'expected None but got %s' % value

    def test_source_to_event(self):

        def stype(x):
            return x.__class__.__name__

        for S in gf.source_classes:
            if S is gf.CombiSource:
                continue

            stf = gf.TriangularSTF(effective_duration=2.0)
            s1 = S(lat=10., lon=20., depth=1000.,
                   north_shift=500., east_shift=500., stf=stf)
            ev = s1.pyrocko_event()

            try:
                s1_mag = s1.get_magnitude()
                mess = ''
            except (gf.DerivedMagnitudeError, NotImplementedError) as e:
                s1_mag = 'N/A'
                mess = '* %s *' % e

            if ev.magnitude is not None:
                ev_mag = ev.magnitude
            else:
                ev_mag = 'N/A'

            s2 = S.from_pyrocko_event(ev)

            if ev.moment_tensor:
                mt_mag = ev.moment_tensor.magnitude
            else:
                mt_mag = 'N/A'

            try:
                s2_mag = s2.get_magnitude()
                mess = ''
            except (gf.DerivedMagnitudeError, NotImplementedError) as e:
                s2_mag = 'N/A'
                mess = '* %s *' % e

            # print(
            #     stype(s1).ljust(32),
            #     s1_mag,
            #     ev.magnitude or 'N/A',
            #     mt_mag,
            #     s2_mag,
            #     mess)

            del mess

            def assert_mag_eq(mag1, mag2):
                if 'N/A' not in (mag1, mag2):
                    num.testing.assert_approx_equal(mag1, mag2)

            assert_mag_eq(s1_mag, ev_mag)
            assert_mag_eq(s1_mag, mt_mag)
            assert_mag_eq(s1_mag, s2_mag)

            if not isinstance(s1, gf.DoubleDCSource):
                assert numeq(
                    [s1.effective_lat, s1.effective_lon,
                     s1.depth, s1.stf.effective_duration],
                    [s2.effective_lat, s2.effective_lon,
                     s2.depth, s2.stf.effective_duration], 0.001)
            else:
                assert numeq(
                    [s1.effective_lat, s1.effective_lon,
                     s1.depth, s1.stf.effective_duration],
                    [s2.effective_lat, s2.effective_lon,
                     s2.depth, s2.stf1.effective_duration], 0.001)

    def test_source_dict(self):
        s1 = gf.DCSource(strike=0.)
        s1.strike = 10.
        s2 = s1.clone(strike=20.)
        s2.update(strike=30.)
        s2['strike'] = 40.
        d = dict(s2)
        s3 = gf.DCSource(**d)
        s3.strike

    def test_sgrid(self):

        r = gf.Range

        source = gf.DCSource()
        sgrid = source.grid(rake=r(-10, 10, 1),
                            strike=r(-100, 100, n=21),
                            depth=r('0k .. 100k : 10k'),
                            magnitude=r(1, 2, 1))

        sgrid = guts.load_string(sgrid.dump())

        n = len(sgrid)
        i = 0
        for source in sgrid:
            i += 1

        assert i == n

    def test_sgrid2(self):
        expect = [10., 12., 14., 16., 18., 20.]
        source = gf.DCSource()
        sgrid = source.grid(dip=gf.Range(10, 20, 2))
        dips = []
        for source in sgrid:
            dips.append(source.dip)

        num.testing.assert_array_almost_equal(
            dips, expect)

        source = gf.DCSource(dip=10)
        sgrid = source.grid(dip=gf.Range(1, 2, 0.2, relative='mult'))
        dips = []
        for source in sgrid:
            dips.append(source.dip)

        num.testing.assert_array_almost_equal(
            dips, expect)

    def dummy_store(self):
        if self._dummy_store is None:

            conf = gf.ConfigTypeA(
                id='empty_regional',
                source_depth_min=0.,
                source_depth_max=20*km,
                source_depth_delta=10*km,
                distance_min=1000*km,
                distance_max=2000*km,
                distance_delta=10*km,
                sample_rate=2.0,
                ncomponents=10,
                earthmodel_1d=cake.load_model(crust2_profile=(50., 10.)))

            store_dir = mkdtemp(prefix='gfstore')
            self.tempdirs.append(store_dir)

            gf.Store.create(store_dir, config=conf)
            self._dummy_store = gf.Store(store_dir)

        return self._dummy_store

    def dummy_homogeneous_store(self):
        mod = cake.LayeredModel.from_scanlines(
            cake.read_nd_model_str('''
0 6 3.46  3.0  1000 500
20 6 3.46  3.0  1000 500
'''.lstrip()))

        if self._dummy_homogeneous_store is None:

            conf = gf.ConfigTypeA(
                id='empty_homogeneous',
                source_depth_min=0.,
                source_depth_max=20*km,
                source_depth_delta=10*km,
                distance_min=1000*km,
                distance_max=2000*km,
                distance_delta=10*km,
                sample_rate=2.0,
                ncomponents=10,
                earthmodel_1d=mod)

            store_dir = mkdtemp(prefix='gfstore')
            self.tempdirs.append(store_dir)

            gf.Store.create(store_dir, config=conf)
            self._dummy_homogeneous_store = gf.Store(store_dir)

        return self._dummy_homogeneous_store

    def test_combine_dsources(self):
        store = self.dummy_store()
        dummy_target = gf.Target()
        for S in gf.source_classes:
            if not hasattr(S, 'discretize_basesource'):
                continue

            for lats in [[10., 10., 10.], [10., 11., 12.]]:
                sources = [
                    default_source(
                        S,
                        lat=lat, lon=20., depth=1000.,
                        north_shift=500., east_shift=500.)

                    for lat in lats]

                dsources = [
                    s.discretize_basesource(store, target=dummy_target)
                    for s in sources]

                DS = dsources[0].__class__

                dsource = DS.combine(dsources)
                assert dsource.nelements == sum(s.nelements for s in dsources)

    def test_source_times(self):
        store = self.dummy_store()
        dummy_target = gf.Target()
        for S in gf.source_classes:
            if not hasattr(S, 'discretize_basesource'):
                continue

            for t in [0.0, util.str_to_time('2014-01-01 10:00:00')]:
                source = default_source(S, time=t)
                dsource = source.discretize_basesource(
                    store, target=dummy_target)
                cent = dsource.centroid()
                assert numeq(cent.time, t, 0.0001)

    def test_outline(self):
        s = gf.MTSource(
            east_shift=5. * km,
            north_shift=-3. * km,
            depth=7. * km)

        outline = s.outline()
        numeq(outline, num.array([[-3000., 5000., 7000.]]), 1e-8)

        rs = gf.RectangularSource(
            length=2 * km,
            width=2 * km)

        outline = rs.outline()
        numeq(
            outline,
            num.array(
                [[-1.e3, 0.0, 0.0],
                 [1.e3, 0.0, 0.0],
                 [1.e3, 0.0, 2.e3],
                 [-1.e3, 0.0, 2.e3],
                 [-1.e3, 0.0, 0.0]]),
            1e-8)

    def test_rect_source_anchors(self):
        sources = {}
        for anchor in ['top', 'center', 'bottom']:
            sources[anchor] = gf.RectangularSource(
                width=5 * km,
                length=20 * km,
                anchor=anchor,
                dip=120.,
                strike=45.,
                east_shift=0 * km,
                north_shift=0 * km)

        def plot_sources(sources):
            import matplotlib.pyplot as plt
            fig = plt.figure()
            ax = fig.gca()
            colors = ['b', 'r', 'y']
            for i, src in enumerate(sources.itervalues()):
                n, e = src.outline(cs='xy').T
                ax.fill(e, n, color=colors[i], alpha=0.5)
            plt.show()

        # plot_sources(sources)

    def test_explosion_source(self):

        target = gf.Target(
            interpolation='nearest_neighbor')

        ex = gf.ExplosionSource(
            magnitude=5.,
            volume_change=4.,
            depth=5 * km)

        with self.assertRaises(gf.DerivedMagnitudeError):
            ex.validate()

        ex = gf.ExplosionSource(
            depth=5 * km)

        ex.validate()

        self.assertEqual(ex.get_moment(), 1.0)

        # magnitude input
        magnitude = 3.
        ex = gf.ExplosionSource(
            magnitude=magnitude,
            depth=5 * km)

        store = self.dummy_store()

        with self.assertRaises(gf.DerivedMagnitudeError):
            ex.get_volume_change()

        volume_change = ex.get_volume_change(
            store, target)

        self.assertAlmostEqual(
            ex.get_magnitude(store, target), magnitude)

        # validate with MT source
        moment = ex.get_moment(store, target) * float(num.sqrt(2. / 3))

        mt = gf.MTSource(mnn=moment, mee=moment, mdd=moment)

        self.assertAlmostEqual(
            ex.get_magnitude(store, target),
            mt.get_magnitude(store=store, target=target))

        # discretized sources
        d_ex = ex.discretize_basesource(store=store, target=target)
        d_mt = mt.discretize_basesource(store=store, target=target)

        d_ex_m6s = d_ex.get_source_terms('elastic10')
        d_mt_m6s = d_mt.get_source_terms('elastic10')

        numeq(d_ex_m6s, d_mt_m6s, 1e-20)

        # interpolation method
        with self.assertRaises(TypeError):
            ex.get_volume_change(
                store, gf.Target(interpolation='nearest_neighbour'))

        # volume change input
        ex = gf.ExplosionSource(
            volume_change=volume_change,
            depth=5 * km)

        self.assertAlmostEqual(
            ex.get_magnitude(store, target), 3.0)

        ex = gf.ExplosionSource(
            magnitude=3.0,
            depth=-5.)

        with self.assertRaises(gf.DerivedMagnitudeError):
            ex.get_volume_change(store, target)

    def test_rect_source(self):

        store = self.dummy_homogeneous_store()

        rect1 = gf.RectangularSource(
            depth=10*km,
            magnitude=5.0,
            width=5*km,
            length=5*km)

        rect2 = gf.RectangularSource(
            depth=10*km,
            slip=pmt.magnitude_to_moment(5.0) / (
                5*km * 5*km * store.config.earthmodel_1d.material(
                    10*km).shear_modulus()),
            width=5*km,
            length=5*km)

        self.assertAlmostEqual(
            rect1.get_magnitude(),
            rect2.get_magnitude(
                store, gf.Target(interpolation='nearest_neighbor')))

    def test_discretize_rect_source(self):

        store = self.dummy_homogeneous_store()
        target = gf.Target(interpolation='nearest_neighbor')

        for source in [
                gf.RectangularSource(
                    depth=10*km,
                    slip=0.5,
                    width=5*km,
                    length=5*km),
                gf.RectangularSource(
                    depth=10*km,
                    magnitude=5.0,
                    width=5*km,
                    length=5*km,
                    decimation_factor=2)]:

            dsource = source.discretize_basesource(store, target)
            m1 = source.get_moment(store, target)
            m2 = dsource.centroid().pyrocko_moment_tensor().scalar_moment()
            assert abs(m1 - m2) < abs(m1 + m2) * 1e-6


if __name__ == '__main__':
    util.setup_logging('test_gf_sources', 'warning')
    unittest.main()



# http://pyrocko.org - GPLv3
#
# The Pyrocko Developers, 21st Century
# ---|P------/S----------~Lg----------
from __future__ import absolute_import
from builtins import object

import calendar
import math
import time
import copy
import logging
import sys

import numpy as num

from pyrocko import util, plot, model, trace
from pyrocko.util import TableWriter, TableReader, gmtime_x, mystrftime


logger = logging.getLogger('pyrocko.gui.marker')


if sys.version_info[0] >= 3:
    polarity_symbols = {1: u'\u2191', -1: u'\u2193', None: u'', 0: u'\u2195'}
else:
    polarity_symbols = {1: '+', -1: '-', None: '', 0: '0'}


def str_to_float_or_none(s):
    if s == 'None':
        return None
    return float(s)


def str_to_str_or_none(s):
    if s == 'None':
        return None
    return s


def str_to_int_or_none(s):
    if s == 'None':
        return None
    return int(s)


def str_to_bool(s):
    return s.lower() in ('true', 't', '1')


def myctime(timestamp):
    tt, ms = gmtime_x(timestamp)
    return mystrftime(None, tt, ms)


class MarkerParseError(Exception):
    pass


class MarkerOneNSLCRequired(Exception):
    pass


class Marker(object):
    '''
    General purpose marker GUI element and base class for
    :py:class:`EventMarker` and :py:class:`PhaseMarker`.

    :param nslc_ids: list of (network, station, location, channel) tuples
        (may contain wildcards)
    :param tmin: start time
    :param tmax: end time
    :param kind: (optional) integer to distinguish groups of markers
        (color-coded)
    '''

    @staticmethod
    def from_string(line):

        def fail():
            raise MarkerParseError(
                'Unable to create marker from string: "%s"' % line)

        def parsedate(ymd, hms, sfs):
            return calendar.timegm(
                time.strptime(ymd+' '+hms, '%Y-%m-%d %H:%M:%S')) + float(sfs)

        try:
            toks = line.split()
            if len(toks) in (4, 5):
                tmin = parsedate(*toks[:3])
                tmax = tmin

            elif len(toks) in (8, 9):
                tmin = parsedate(*toks[:3])
                tmax = parsedate(*toks[3:6])

            else:
                fail()

            if len(toks) in (5, 9):
                kind = int(toks[-2])
            else:
                kind = int(toks[-1])

            nslc_ids = []
            if len(toks) in (5, 9):
                for snslc in toks[-1].split(','):
                    nslc = snslc.split('.')
                    if len(nslc) != 4:
                        fail()

                    nslc_ids.append(tuple(nslc))

        except MarkerParseError:
            fail()

        return Marker(nslc_ids, tmin, tmax, kind=kind)

    @staticmethod
    def save_markers(markers, fn, fdigits=3):
        '''Static method to write marker objects to file.

        :param markers: list of :py:class:`Marker` objects
        :param fn: filename as string
        :param fdigits: number of decimal digits to use for sub-second time
             strings (default 3)
        '''
        f = open(fn, 'w')
        f.write('# Snuffler Markers File Version 0.2\n')
        writer = TableWriter(f)
        for marker in markers:
            a = marker.get_attributes(fdigits=fdigits)
            w = marker.get_attribute_widths(fdigits=fdigits)
            row = []
            for x in a:
                if x is None or x == '':
                    row.append('None')
                else:
                    row.append(x)

            writer.writerow(row, w)

        f.close()

    @staticmethod
    def load_markers(fn):
        '''
        Static method to load markers from file.

        :param filename:  filename as string
        :returns: list of :py:class:`Marker`, :py:class:`EventMarker` or
            :py:class:`PhaseMarker` objects
        '''
        markers = []
        with open(fn, 'r') as f:
            line = f.readline()
            if not line.startswith('# Snuffler Markers File Version'):
                f.seek(0)
                for iline, line in enumerate(f):
                    line = str(line.decode('ascii'))
                    sline = line.strip()
                    if not sline or sline.startswith('#'):
                        continue
                    try:
                        m = Marker.from_string(sline)
                        markers.append(m)

                    except MarkerParseError:
                        logger.warning(
                            'Invalid marker definition in line %i of file "%s"'
                            % (iline+1, fn))

                f.close()

            elif line.startswith('# Snuffler Markers File Version 0.2'):
                reader = TableReader(f)
                while not reader.eof:
                    row = reader.readrow()
                    if not row:
                        continue
                    if row[0] == 'event:':
                        marker = EventMarker.from_attributes(row)
                    elif row[0] == 'phase:':
                        marker = PhaseMarker.from_attributes(row)
                    else:
                        marker = Marker.from_attributes(row)

                    markers.append(marker)
            else:
                logger.warning('Unsupported Markers File Version')

        return markers

    def __init__(self, nslc_ids, tmin, tmax, kind=0):
        self.set(nslc_ids, tmin, tmax)
        c = plot.color

        self.color_a = [c(x) for x in (
            'aluminium4', 'aluminium5', 'aluminium6')]

        self.color_b = [c(x) for x in (
            'scarletred1', 'scarletred2', 'scarletred3',
            'chameleon1', 'chameleon2', 'chameleon3',
            'skyblue1', 'skyblue2', 'skyblue3',
            'orange1', 'orange2', 'orange3',
            'plum1', 'plum2', 'plum3',
            'chocolate1', 'chocolate2', 'chocolate3')]

        self.alerted = False
        self.selected = False
        self.kind = kind

    def set(self, nslc_ids, tmin, tmax):
        '''Set ``nslc_ids``, start time and end time of :py:class:`Marker`

        :param nslc_ids: list or set of (network, station, location, channel)
            tuples
        :param tmin: start time
        :param tmax: end time
        '''
        self.nslc_ids = nslc_ids
        self.tmin = tmin
        self.tmax = tmax

    def set_kind(self, kind):
        '''Set kind of :py:class:`Marker`

        :param kind: (optional) integer to distinguish groups of markers
                        (color-coded)
        '''
        self.kind = kind

    def get_tmin(self):
        '''Get *start time* of :py:class:`Marker`'''
        return self.tmin

    def get_tmax(self):
        '''Get *end time* of :py:class:`Marker`'''
        return self.tmax

    def get_nslc_ids(self):
        '''Get marker's network-station-location-channel pattern.

        :returns: list or set of (network, station, location, channel) tuples

        The network, station, location, or channel strings may contain wildcard
        expressions.
        '''
        return self.nslc_ids

    def is_alerted(self):
        return self.alerted

    def is_selected(self):
        return self.selected

    def set_alerted(self, state):
        self.alerted = state

    def set_selected(self, state):
        self.selected = state

    def match_nsl(self, nsl):
        '''See documentation of :py:func:`pyrocko.util.match_nslc`'''
        patterns = ['.'.join(x[:3]) for x in self.nslc_ids]
        return util.match_nslc(patterns, nsl)

    def match_nslc(self, nslc):
        '''See documentation of :py:func:`pyrocko.util.match_nslc`'''
        patterns = ['.'.join(x) for x in self.nslc_ids]
        return util.match_nslc(patterns, nslc)

    def one_nslc(self):
        '''If one *nslc_id* defines this marker return this id.
        If more than one *nslc_id* is defined in the :py:class:`Marker`s
        *nslc_ids* raise :py:exc:`MarkerOneNSLCRequired`.'''
        if len(self.nslc_ids) != 1:
            raise MarkerOneNSLCRequired()

        return list(self.nslc_ids)[0]

    def hoover_message(self):
        return ''

    def copy(self):
        ''' Get a copy of this marker. '''
        return copy.deepcopy(self)

    def __str__(self):
        traces = ','.join(['.'.join(nslc_id) for nslc_id in self.nslc_ids])
        st = myctime
        if self.tmin == self.tmax:
            return '%s %i %s' % (st(self.tmin), self.kind, traces)
        else:
            return '%s %s %g %i %s' % (
                st(self.tmin), st(self.tmax), self.tmax-self.tmin, self.kind,
                traces)

    def get_attributes(self, fdigits=3):
        traces = ','.join(['.'.join(nslc_id) for nslc_id in self.nslc_ids])

        def st(t):
            return util.time_to_str(
                t, format='%Y-%m-%d %H:%M:%S.'+'%iFRAC' % fdigits)

        vals = []
        vals.extend(st(self.tmin).split())
        if self.tmin != self.tmax:
            vals.extend(st(self.tmax).split())
            vals.append(self.tmax-self.tmin)

        vals.append(self.kind)
        vals.append(traces)
        return vals

    def get_attribute_widths(self, fdigits=3):
        ws = [10, 9+fdigits]
        if self.tmin != self.tmax:
            ws.extend([10, 9+fdigits, 12])
        ws.extend([2, 15])
        return ws

    @staticmethod
    def parse_attributes(vals):
        tmin = util.str_to_time(vals[0] + ' ' + vals[1])
        i = 2
        tmax = tmin
        if len(vals) == 7:
            tmax = util.str_to_time(vals[2] + ' ' + vals[3])
            i = 5

        kind = int(vals[i])
        traces = vals[i+1]
        if traces == 'None':
            nslc_ids = []
        else:
            nslc_ids = tuple(
                [tuple(nslc_id.split('.')) for nslc_id in traces.split(',')])

        return nslc_ids, tmin, tmax, kind

    @staticmethod
    def from_attributes(vals):
        return Marker(*Marker.parse_attributes(vals))

    def select_color(self, colorlist):

        def cl(x):
            return colorlist[(self.kind*3+x) % len(colorlist)]

        if self.selected:
            return cl(1)

        if self.alerted:
            return cl(1)

        return cl(2)

    def draw(
            self, p, time_projection, y_projection,
            draw_line=True,
            draw_triangle=False,
            **kwargs):

        from .qt_compat import qc, qg
        from . import util as gui_util

        if self.selected or self.alerted or not self.nslc_ids:

            color = self.select_color(self.color_b)
            pen = qg.QPen(qg.QColor(*color))
            pen.setWidth(2)
            linepen = qg.QPen(pen)
            if self.selected or self.alerted:
                linepen.setStyle(qc.Qt.CustomDashLine)
                pat = [5., 3.]
                linepen.setDashPattern(pat)
                if self.alerted and not self.selected:
                    linepen.setColor(qg.QColor(150, 150, 150))

            s = 9.
            utriangle = gui_util.make_QPolygonF(
                [-0.577*s, 0., 0.577*s], [0., 1.*s, 0.])
            ltriangle = gui_util.make_QPolygonF(
                [-0.577*s, 0., 0.577*s], [0., -1.*s, 0.])

            def drawline(t):
                u = time_projection(t)
                v0, v1 = y_projection.get_out_range()
                line = qc.QLineF(u, v0, u, v1)
                p.drawLine(line)

            def drawtriangles(t):
                u = time_projection(t)
                v0, v1 = y_projection.get_out_range()
                t = qg.QPolygonF(utriangle)
                t.translate(u, v0)
                p.drawConvexPolygon(t)
                t = qg.QPolygonF(ltriangle)
                t.translate(u, v1)
                p.drawConvexPolygon(t)

            if draw_line or self.selected or self.alerted:
                p.setPen(linepen)
                drawline(self.tmin)
                drawline(self.tmax)

            if draw_triangle:
                pen.setStyle(qc.Qt.SolidLine)
                pen.setJoinStyle(qc.Qt.MiterJoin)
                pen.setWidth(2)
                p.setPen(pen)
                p.setBrush(qg.QColor(*color))
                drawtriangles(self.tmin)

    def draw_trace(
            self, viewer, p, tr, time_projection, track_projection, gain,
            outline_label=False):

        from .qt_compat import qc, qg
        from . import util as gui_util

        if self.nslc_ids and not self.match_nslc(tr.nslc_id):
            return

        color = self.select_color(self.color_b)
        pen = qg.QPen(qg.QColor(*color))
        pen.setWidth(2)
        p.setPen(pen)
        p.setBrush(qc.Qt.NoBrush)

        def drawpoint(t, y):
            u = time_projection(t)
            v = track_projection(y)
            rect = qc.QRectF(u-2, v-2, 4, 4)
            p.drawRect(rect)

        def drawline(t):
            u = time_projection(t)
            v0, v1 = track_projection.get_out_range()
            line = qc.QLineF(u, v0, u, v1)
            p.drawLine(line)

        try:
            snippet = tr.chop(
                self.tmin, self.tmax,
                inplace=False,
                include_last=True,
                snap=(math.ceil, math.floor))

            vdata = track_projection(gain*snippet.get_ydata())
            udata_min = float(
                time_projection(snippet.tmin))
            udata_max = float(
                time_projection(snippet.tmin+snippet.deltat*(vdata.size-1)))
            udata = num.linspace(udata_min, udata_max, vdata.size)
            qpoints = gui_util.make_QPolygonF(udata, vdata)
            pen.setWidth(1)
            p.setPen(pen)
            p.drawPolyline(qpoints)
            pen.setWidth(2)
            p.setPen(pen)
            drawpoint(*tr(self.tmin, clip=True, snap=math.ceil))
            drawpoint(*tr(self.tmax, clip=True, snap=math.floor))

        except trace.NoData:
            pass

        color = self.select_color(self.color_b)
        pen = qg.QPen(qg.QColor(*color))
        pen.setWidth(2)
        p.setPen(pen)

        drawline(self.tmin)
        drawline(self.tmax)

        label = self.get_label()
        if label:
            label_bg = qg.QBrush(qg.QColor(255, 255, 255))

            u = time_projection(self.tmin)
            v0, v1 = track_projection.get_out_range()
            if outline_label:
                du = -7
            else:
                du = -5
            gui_util.draw_label(
                p, u+du, v0, label, label_bg, 'TR',
                outline=outline_label)

        if self.tmin == self.tmax:
            try:
                drawpoint(self.tmin, tr.interpolate(self.tmin))

            except IndexError:
                pass

    def get_label(self):
        return None

    def convert_to_phase_marker(
            self,
            event=None,
            phasename=None,
            polarity=None,
            automatic=None,
            incidence_angle=None,
            takeoff_angle=None):

        if isinstance(self, PhaseMarker):
            return

        self.__class__ = PhaseMarker
        self._event = event
        self._phasename = phasename
        self._polarity = polarity
        self._automatic = automatic
        self._incidence_angle = incidence_angle
        self._takeoff_angle = takeoff_angle
        if self._event:
            self._event_hash = event.get_hash()
            self._event_time = event.time
        else:
            self._event_hash = None
            self._event_time = None

    def convert_to_event_marker(self, lat=0., lon=0.):
        if isinstance(self, EventMarker):
            return

        if isinstance(self, PhaseMarker):
            self.convert_to_marker()

        self.__class__ = EventMarker
        self._event = model.Event(lat, lon, self.tmin, name='Event')
        self._event_hash = self._event.get_hash()
        self._active = False
        self.tmax = self.tmin
        self.nslc_ids = []


class EventMarker(Marker):
    """
    An EventMarker is a GUI element representing a seismological event

    :param event: A :py:class:`pyrocko.model.Event` object containing meta
        information of a seismological event
    :param kind: (optional) integer to distinguish groups of markers
    :param event_hash:  (optional) hash code of event (see:
        :py:meth:`pyrocko.model.Event.get_hash`)
    """
    def __init__(self, event, kind=0, event_hash=None):
        Marker.__init__(self, [], event.time, event.time, kind)
        self._event = event
        self._active = False
        self._event_hash = event_hash

    def get_event_hash(self):
        if self._event_hash is not None:
            return self._event_hash
        else:
            return self._event.get_hash()

    def set_active(self, active):
        self._active = active

    def label(self):
        t = []
        mag = self._event.magnitude
        if mag is not None:
            t.append('M%3.1f' % mag)

        reg = self._event.region
        if reg is not None:
            t.append(reg)

        nam = self._event.name
        if nam is not None:
            t.append(nam)

        s = ' '.join(t)
        if not s:
            s = '(Event)'
        return s

    def draw(self, p, time_projection, y_projection, with_label=False):
        Marker.draw(
            self, p, time_projection, y_projection,
            draw_line=False,
            draw_triangle=True)

        if with_label:
            self.draw_label(p, time_projection, y_projection)

    def draw_label(self, p, time_projection, y_projection):
        from .qt_compat import qg
        from . import util as gui_util

        u = time_projection(self.tmin)
        v0, v1 = y_projection.get_out_range()
        label_bg = qg.QBrush(qg.QColor(255, 255, 255))
        gui_util.draw_label(
            p, u, v0-10., self.label(), label_bg, 'CB',
            outline=self._active)

    def get_event(self):
        '''Return an instance of the :py:class:`pyrocko.model.Event` associated
        to this :py:class:`EventMarker`'''
        return self._event

    def draw_trace(self, viewer, p, tr, time_projection, track_projection,
                   gain):
        pass

    def hoover_message(self):
        ev = self.get_event()
        evs = []
        for k in 'magnitude lat lon depth name region catalog'.split():
            if ev.__dict__[k] is not None and ev.__dict__[k] != '':
                if k == 'depth':
                    sv = '%g km' % (ev.depth * 0.001)
                else:
                    sv = '%s' % ev.__dict__[k]
                evs.append('%s = %s' % (k, sv))

        return ', '.join(evs)

    def get_attributes(self, fdigits=3):
        attributes = ['event:']
        attributes.extend(Marker.get_attributes(self, fdigits=fdigits))
        del attributes[-1]
        e = self._event
        attributes.extend([
            e.get_hash(), e.lat, e.lon, e.depth, e.magnitude, e.catalog,
            e.name, e.region])

        return attributes

    def get_attribute_widths(self, fdigits=3):
        ws = [6]
        ws.extend(Marker.get_attribute_widths(self, fdigits=fdigits))
        del ws[-1]
        ws.extend([14, 12, 12, 12, 4, 5, 0, 0])
        return ws

    @staticmethod
    def from_attributes(vals):

        nslc_ids, tmin, tmax, kind = Marker.parse_attributes(
            vals[1:] + ['None'])
        lat, lon, depth, magnitude = [
            str_to_float_or_none(x) for x in vals[5:9]]
        catalog, name, region = [
            str_to_str_or_none(x) for x in vals[9:]]
        e = model.Event(
            lat, lon, tmin, name, depth, magnitude, region=region, catalog=catalog)
        marker = EventMarker(
            e, kind, event_hash=str_to_str_or_none(vals[4]))
        return marker


class PhaseMarker(Marker):
    '''
    A PhaseMarker is a GUI-element representing a seismological phase arrival

    :param nslc_ids: list of (network, station, location, channel) tuples (may
        contain wildcards)
    :param tmin: start time
    :param tmax: end time
    :param kind: (optional) integer to distinguish groups of markers
        (color-coded)
    :param event: a :py:class:`pyrocko.model.Event` object containing meta
        information of a seismological event
    :param event_hash: (optional) hash code of event (see:
        :py:meth:`pyrocko.model.Event.get_hash`)
    :param event_time: (optional) time of the associated event
    :param phasename: (optional) name of the phase associated with the marker
    :param polarity: (optional) polarity of arriving phase
    :param automatic: (optional)
    :param incident_angle: (optional) incident angle of phase
    :param takeoff_angle: (optional) take off angle of phase
    '''
    def __init__(
            self, nslc_ids, tmin, tmax,
            kind=0,
            event=None,
            event_hash=None,
            event_time=None,
            phasename=None,
            polarity=None,
            automatic=None,
            incidence_angle=None,
            takeoff_angle=None):

        Marker.__init__(self, nslc_ids, tmin, tmax, kind)
        self._event = event
        self._event_hash = event_hash
        self._event_time = event_time
        self._phasename = phasename
        self._automatic = automatic
        self._incidence_angle = incidence_angle
        self._takeoff_angle = takeoff_angle

        self.set_polarity(polarity)

    def draw_trace(self, viewer, p, tr, time_projection, track_projection,
                   gain):

        Marker.draw_trace(
            self, viewer, p, tr, time_projection, track_projection, gain,
            outline_label=(
                self._event is not None and
                self._event == viewer.get_active_event()))

    def get_label(self):
        t = []
        if self._phasename is not None:
            t.append(self._phasename)
        if self._polarity is not None:
            t.append(self.get_polarity_symbol())

        if self._automatic:
            t.append('@')

        return ''.join(t)

    def get_event(self):
        '''Return an instance of the :py:class:`pyrocko.model.Event` associated
        to this :py:class:`EventMarker`'''
        return self._event

    def get_event_hash(self):
        if self._event_hash is not None:
            return self._event_hash
        else:
            if self._event is None:
                return None
            else:
                return self._event.get_hash()

    def get_event_time(self):
        if self._event is not None:
            return self._event.time
        else:
            return self._event_time

    def set_event_hash(self, event_hash):
        self._event_hash = event_hash

    def set_event(self, event):
        self._event = event
        if event is not None:
            self.set_event_hash(event.get_hash())

    def get_phasename(self):
        return self._phasename

    def set_phasename(self, phasename):
        self._phasename = phasename

    def set_polarity(self, polarity):
        if polarity not in [1, -1, 0, None]:
            raise ValueError('polarity has to be 1, -1, 0 or None')
        self._polarity = polarity

    def get_polarity_symbol(self):
        return polarity_symbols.get(self._polarity, '')

    def get_polarity(self):
        return self._polarity

    def convert_to_marker(self):
        del self._event
        del self._event_hash
        del self._phasename
        del self._polarity
        del self._automatic
        del self._incidence_angle
        del self._takeoff_angle
        self.__class__ = Marker

    def hoover_message(self):
        toks = []
        for k in 'incidence_angle takeoff_angle polarity'.split():
            v = getattr(self, '_' + k)
            if v is not None:
                toks.append('%s = %s' % (k, v))

        return ', '.join(toks)

    def get_attributes(self, fdigits=3):
        attributes = ['phase:']
        attributes.extend(Marker.get_attributes(self, fdigits=fdigits))

        et = None, None
        if self._event:
            et = self._st(self._event.time, fdigits).split()
        elif self._event_time:
            et = self._st(self._event_time, fdigits).split()

        attributes.extend([
            self.get_event_hash(), et[0], et[1], self._phasename,
            self._polarity, self._automatic])

        return attributes

    def _st(self, t, fdigits):
        return util.time_to_str(
            t, format='%Y-%m-%d %H:%M:%S.'+'%iFRAC' % fdigits)

    def get_attribute_widths(self, fdigits=3):
        ws = [6]
        ws.extend(Marker.get_attribute_widths(self, fdigits=fdigits))
        ws.extend([14, 12, 12, 8, 4, 5])
        return ws

    @staticmethod
    def from_attributes(vals):
        if len(vals) == 14:
            nbasicvals = 7
        else:
            nbasicvals = 4
        nslc_ids, tmin, tmax, kind = Marker.parse_attributes(
            vals[1:1+nbasicvals])

        i = 8
        if len(vals) == 14:
            i = 11

        event_hash = str_to_str_or_none(vals[i-3])
        event_sdate = str_to_str_or_none(vals[i-2])
        event_stime = str_to_str_or_none(vals[i-1])

        if event_sdate is not None and event_stime is not None:
            event_time = util.str_to_time(event_sdate + ' ' + event_stime)
        else:
            event_time = None

        phasename = str_to_str_or_none(vals[i])
        polarity = str_to_int_or_none(vals[i+1])
        automatic = str_to_bool(vals[i+2])
        marker = PhaseMarker(nslc_ids, tmin, tmax, kind, event=None,
                             event_hash=event_hash, event_time=event_time,
                             phasename=phasename, polarity=polarity,
                             automatic=automatic)
        return marker


def load_markers(filename):
    '''
    Load markers from file.

    :param filename:  filename as string
    :returns: list of :py:class:`Marker` Objects
    '''

    return Marker.load_markers(filename)


def save_markers(markers, filename, fdigits=3):
    '''
    Save markers to file.

    :param markers: list of :py:class:`Marker` Objects
    :param filename: filename as string
    :param fdigits: number of decimal digits to use for sub-second time strings
    '''

    return Marker.save_markers(markers, filename, fdigits=fdigits)


def associate_phases_to_events(markers):
    '''
    Reassociate phases to events after import from markers file.
    '''

    hash_to_events = {}
    time_to_events = {}
    for marker in markers:
        if isinstance(marker, EventMarker):
            ev = marker.get_event()
            hash_to_events[marker.get_event_hash()] = ev
            time_to_events[ev.time] = ev

    for marker in markers:
        if isinstance(marker, PhaseMarker):
            h = marker.get_event_hash()
            t = marker.get_event_time()
            if marker.get_event() is None:
                if h is not None and h in hash_to_events:
                    marker.set_event(hash_to_events[h])
                    marker.set_event_hash(None)
                elif t is not None and t in time_to_events:
                    marker.set_event(time_to_events[t])
                    marker.set_event_hash(None)



from __future__ import division, print_function, absolute_import
import os
import math
import tempfile
import shutil
import unittest
import numpy as num
from matplotlib import image, pyplot as plt

from pyrocko import util
from pyrocko.plot import automap, gmtpy

from . import common

noshow = False
km = 1000.


@unittest.skipUnless(
    gmtpy.have_gmt(), 'GMT not available')
@unittest.skipUnless(
    gmtpy.have_pixmaptools(), '`pdftocairo` or `convert` not available')
class AutomapTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def fpath(self, fn):
        return os.path.join(self.tempdir, fn)

    def fpath_ref(self, fn):
        try:
            return common.test_data_file(fn)
        except util.DownloadError:
            return common.test_data_file_no_download(fn)

    def compare_with_ref(self, fname, tolerance, show=False):
        fpath = self.fpath(fname)
        fpath_ref = self.fpath_ref(fname)

        if not os.path.exists(fpath_ref):
            shutil.copy(fpath, fpath_ref)

        img = image.imread(fpath)
        img_ref = image.imread(fpath_ref)
        self.assertEqual(img.shape, img_ref.shape)
        d = num.abs(img - img_ref)
        merr = num.mean(d)
        if (merr > tolerance or show) and not noshow:
            fig = plt.figure()
            axes1 = fig.add_subplot(1, 3, 1, aspect=1.)
            axes2 = fig.add_subplot(1, 3, 2, aspect=1.)
            axes3 = fig.add_subplot(1, 3, 3, aspect=1.)
            axes1.imshow(img)
            axes1.set_title('Candidate')
            axes2.imshow(img_ref)
            axes2.set_title('Reference')
            axes3.imshow(d)
            axes3.set_title('Mean abs difference: %g' % merr)
            plt.show()
            plt.close(fig)

        assert merr <= tolerance

    def test_napoli(self):
        for version in gmtpy.all_installed_gmt_versions():
            m = automap.Map(
                gmtversion=version,
                lat=40.85,
                lon=14.27,
                radius=50.*km,
                width=20.,
                height=20.,
                margins=[2., 2., 2., 4.5],
                show_grid=True,
                show_topo=True,
                show_topo_scale=True,
                illuminate=True,
                custom_cities=[
                    automap.City(
                        'Pompeji', 40.7506, 14.4897, population=9000)],
                comment='This is output from test_napoli!')

            assert m.have_coastlines()

            m.draw_cities(include=['Pompeji', 'Capri'])
            fname = 'automap_test_napoli.png'
            fpath = self.fpath(fname)
            m.save(fpath)
            self.compare_with_ref(fname, 0.01, show=False)

    def test_fidshi(self):
        x = num.linspace(178., 182., 101)
        y = num.linspace(-17.5, -14.5, 101)
        z = num.sin((x[num.newaxis, :]-x[0])/(x[-1]-x[0])*2.*math.pi*5) * \
            num.sin((y[:, num.newaxis]-y[0])/(y[-1]-y[0])*2.*math.pi*5) * 5000.

        tile = automap.FloatTile(
            xmin=x[0],
            ymin=y[0],
            dx=x[1] - x[0],
            dy=y[1] - y[0],
            data=z)

        for version in gmtpy.all_installed_gmt_versions():
            m = automap.Map(
                gmtversion=version,
                lat=-16.5,
                lon=180.,
                radius=100.*km,
                width=20.,
                height=20.,
                margins=[2., 2., 2., 4.5],
                show_grid=True,
                show_topo=True,
                show_topo_scale=True,
                replace_topo_color_only=tile,
                illuminate=True)

            m.draw_cities()
            fname = 'automap_test_fidshi.png'
            fpath = self.fpath(fname)
            m.save(fpath)
            self.compare_with_ref(fname, 0.01, show=False)

    def test_new_zealand(self):
        m = automap.Map(
            lat=-42.57,
            lon=173.01,
            radius=1000.*km,
            width=20.,
            height=20.,
            color_dry=gmtpy.color_tup('aluminium1'),
            show_topo=False,
            show_rivers=False,
            show_plates=True)

        m.draw_cities()
        fname = 'new_zealand.pdf'
        fpath = self.fpath(fname)
        m.save(fpath)


if __name__ == "__main__":
    util.setup_logging('test_automap', 'warning')
    unittest.main()



import numpy as num
import logging
import sys
import os
import os.path as op

from pyrocko.guts import List
from pyrocko.plot import gmtpy
from pyrocko.gui.marker import PhaseMarker
from pyrocko import pile, util, model
from pyrocko.dataset import topo

from .base import LocationGenerator, ScenarioError
from .sources import SourceGenerator, DCSourceGenerator
from .targets import TargetGenerator, AVAILABLE_TARGETS

logger = logging.getLogger('pyrocko.scenario')
guts_prefix = 'pf.scenario'


class CannotCreate(Exception):
    pass


class ScenarioGenerator(LocationGenerator):

    target_generators = List.T(
        TargetGenerator.T(),
        default=[],
        help='Targets to spawn in the scenario.')

    source_generator = SourceGenerator.T(
        default=DCSourceGenerator.D(),
        help='Sources to spawn in the scenario.')

    def __init__(self, **kwargs):
        LocationGenerator.__init__(self, **kwargs)

        for gen in self.target_generators:
            gen.update_hierarchy(self)

        for itry in range(self.ntries):

            try:
                self.get_stations()
                self.get_sources()
                return

            except ScenarioError:
                self.retry()

        raise ScenarioError(
            'could not generate scenario within %i tries' % self.ntries)

    def init_modelling(self, engine):
        self._engine = engine

    def get_engine(self):
        return self._engine

    def get_sources(self):
        return self.source_generator.get_sources()

    def get_events(self):
        return [s.pyrocko_event() for s in self.get_sources()]

    def collect(collector):
        if not callable(collector):
            raise AttributeError('This method should not be called directly.')

        def method(self, *args, **kwargs):
            result = []
            func = collector(self, *args, **kwargs)
            for gen in self.target_generators:
                result.extend(
                    func(gen))
            return result

        return method

    @collect
    def get_stations(self):
        return lambda gen: gen.get_stations()

    @collect
    def get_waveforms(self, tmin=None, tmax=None):
        tmin, tmax = self._time_range_fill_defaults(tmin, tmax)
        return lambda gen: gen.get_waveforms(
            self._engine, self.get_sources(),
            tmin=tmin, tmax=tmax)

    @collect
    def get_onsets(self, tmin=None, tmax=None):
        tmin, tmax = self._time_range_fill_defaults(tmin, tmax)
        return lambda gen: gen.get_onsets(
            self._engine, self.get_sources(),
            tmin=tmin, tmax=tmax)

    @collect
    def get_insar_scenes(self, tmin=None, tmax=None):
        tmin, tmax = self._time_range_fill_defaults(tmin, tmax)
        return lambda gen: gen.get_insar_scenes(
            self._engine, self.get_sources(),
            tmin=tmin, tmax=tmax)

    @collect
    def get_gnss_campaigns(self, tmin=None, tmax=None):
        tmin, tmax = self._time_range_fill_defaults(tmin, tmax)
        return lambda gen: gen.get_gnss_campaigns(
            self._engine, self.get_sources(),
            tmin=tmin, tmax=tmax)

    @collect
    def dump_data(self, path, tmin=None, tmax=None, overwrite=False):
        tmin, tmax = self._time_range_fill_defaults(tmin, tmax)
        self.source_generator.dump_data(path)

        meta_dir = op.join(path, 'meta')
        util.ensuredir(meta_dir)

        model.station.dump_stations(
            self.get_stations(), op.join(meta_dir, 'stations.txt'))
        model.station.dump_kml(
            self.get_stations(), op.join(meta_dir, 'stations.kml'))

        markers = self.get_onsets()
        if markers:
            PhaseMarker.save_markers(
                markers, op.join(meta_dir, 'markers.txt'))

        dump_readme(path)

        def dump_data(gen):
            logger.info('Creating files from %s...' % gen.__class__.__name__)
            return gen.dump_data(
                self._engine, self.get_sources(), path,
                tmin=tmin, tmax=tmax, overwrite=overwrite)

        return dump_data

    @collect
    def _get_time_ranges(self):
        return lambda gen: [gen.get_time_range(self.get_sources())]

    def get_time_range(self):
        ranges = num.array(self._get_time_ranges())
        return ranges.min(), ranges.max()

    def _time_range_fill_defaults(self, tmin, tmax):
        stmin, stmax = self.get_time_range()
        return stmin if tmin is None else tmin, stmax if tmax is None else tmax

    def get_pile(self, tmin=None, tmax=None):
        p = pile.Pile()

        trf = pile.MemTracesFile(None, self.get_waveforms(tmin, tmax))
        p.add_file(trf)
        return p

    def make_map(self, filename):
        logger.info('Plotting scenario\'s map...')
        if not gmtpy.have_gmt():
            logger.warning('Cannot plot map, GMT is not installed.')
        try:
            draw_scenario_gmt(self, filename)
        except gmtpy.GMTError:
            logger.warning('GMT threw an error, could not plot map.')
        except topo.AuthenticationRequired:
            logger.warning('Cannot download topography data (authentication '
                           'required). Could not plot map.')

    def draw_map(self, fn):
        from pyrocko.plot import automap

        lat, lon = self.get_center_latlon()
        radius = self.get_radius()

        m = automap.Map(
            width=30.,
            height=30.,
            lat=lat,
            lon=lon,
            radius=radius,
            show_topo=True,
            show_grid=True,
            show_rivers=True,
            color_wet=(216, 242, 254),
            color_dry=(238, 236, 230)
            )

        self.source_generator.add_map_artists(m)

        sources = self.get_sources()
        for gen in self.target_generators:
            gen.add_map_artists(self.get_engine(), sources, m)

        # for patch in self.get_insar_patches():
        #     symbol_size = 50.
        #     coords = num.array(patch.get_corner_coordinates())
        #     m.gmt.psxy(in_rows=num.fliplr(coords),
        #                L=True,
        #                *m.jxyr)

        m.save(fn)

    @property
    def stores_wanted(self):
        return set([gen.store_id for gen in self.target_generators
                    if hasattr(gen, 'store_id')])

    @property
    def stores_missing(self):
        return self.stores_wanted - set(self.get_engine().get_store_ids())

    def ensure_gfstores(self, interactive=False):
        if not self.stores_missing:
            return

        from pyrocko.gf import ws

        engine = self.get_engine()

        gf_store_superdirs = engine.store_superdirs

        if interactive:
            print('We could not find the following Green\'s function stores:\n'
                  '%s\n'
                  'We can try to download the stores from '
                  'http://kinherd.org into one of the following '
                  'directories.'
                  % '\n'.join('  ' + s for s in self.stores_missing))
            for idr, dr in enumerate(gf_store_superdirs):
                print(' %d. %s' % ((idr+1), dr))
            s = input('\nInto which directory should we download the GF '
                      'store(s)?\nDefault 1, (C)ancel: ')
            if s in ['c', 'C']:
                print('Canceled!')
                sys.exit(1)
            elif s == '':
                s = 0
            try:
                s = int(s)
                if s > len(gf_store_superdirs):
                    raise ValueError
            except ValueError:
                print('Invalid selection: %s' % s)
                sys.exit(1)
        else:
            s = 1

        download_dir = gf_store_superdirs[s-1]
        util.ensuredir(download_dir)
        logger.info('Downloading Green\'s functions stores to %s'
                    % download_dir)

        oldwd = os.getcwd()
        for store in self.stores_missing:
            os.chdir(download_dir)
            ws.download_gf_store(site='kinherd', store_id=store)

        os.chdir(oldwd)

    @classmethod
    def initialize(
            cls, path,
            center_lat=None, center_lon=None, radius=None,
            targets=AVAILABLE_TARGETS, force=False):
        """Initialize a Scenario and create a ``scenario.yml``

        :param path: Path to create the scenerio in
        :type path: str
        :param center_lat: Center latitude, defaults to None
        :type center_lat: float, optional
        :param center_lon: Center longitude, defaults to None
        :type center_lon: float, optional
        :param radius: Scenario's radius in [m], defaults to None
        :type radius: float, optional
        :param targets: Targets to thow into scenario,
            defaults to AVAILABLE_TARGETS
        :type targets: list of :class:`pyrocko.scenario.ScenarioTargets`,
            optional
        :param force: Overwrite directory, defaults to False
        :type force: bool, optional
        :returns: Scenario
        :rtype: :class:`pyrocko.scenario.ScenarioGenerator`
        """
        import os.path as op

        if op.exists(path) and not force:
            raise CannotCreate('Directory %s alread exists! May use force?'
                               % path)

        util.ensuredir(path)
        fn = op.join(path, 'scenario.yml')
        logger.debug('Writing new scenario to %s' % fn)

        scenario = cls()
        scenario.target_generators.extend([t() for t in targets])

        for gen in scenario.target_generators:
            gen.update_hierarchy(scenario)

        scenario.center_lat = center_lat
        scenario.center_lon = center_lon
        scenario.radius = radius

        scenario.dump(filename=fn)

        return scenario


def draw_scenario_gmt(generator, fn):
    return generator.draw_map(fn)


def dump_readme(path):
    readme = '''# Pyrocko Earthquake Scenario

The directory structure of a scenario is layed out as follows:

## Map of the scenario
A simple map is generated from `pyrocko.automap` in map.pdf

## Earthquake Sources

Can be found as events.txt and sources.yml hosts the pyrocko.gf sources.

## Folder `meta`

Contains stations.txt and StationXML files for waveforms as well as KML data.
The responses are flat with gain of 1.0 at 1.0 Hz.

## Folder `waveforms`

Waveforms as mini-seed are stored here, segregated into days.

## Folder `gnss`

The GNSS campaign.yml is living here.
Use `pyrocko.guts.load(filename='campaign.yml)` to load the campaign.

## Folder `insar`

Kite InSAR scenes for ascending and descending tracks are stored there.
Use `kite.Scene.load(<filename>)` to inspect the scenes.

'''
    fn = op.join(path, 'README.md')
    with open(fn, 'w') as f:
        f.write(readme)

    return [fn]



from __future__ import division, print_function, absolute_import
import unittest
import shutil
import os
import asyncore
import threading
import logging
import tempfile

from pyrocko.gf import server, LocalEngine, ws, store
from pyrocko import util
from pyrocko.fomosto import ahfullgreen

op = os.path
logger = logging.getLogger('pyrocko.test.test_gf_ws')


class GFWSTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cwd = os.getcwd()

        cls.serve_dir = tempfile.mkdtemp(prefix='pyrocko')

        ahfullgreen.init(cls.serve_dir, None)
        cls.store = store.Store(cls.serve_dir, 'r')
        cls.store.make_ttt()
        cls.store_id = cls.store.config.id
        ahfullgreen.build(cls.serve_dir)

        cls.dl_dir = tempfile.mkdtemp(prefix='pyrocko')
        os.chdir(cls.dl_dir)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.cwd)
        shutil.rmtree(cls.serve_dir)
        shutil.rmtree(cls.dl_dir)

    def test_local_server(self):

        class ServerThread(threading.Thread):
            def __init__(self, serve_dir):
                threading.Thread.__init__(self)
                self.engine = LocalEngine(
                    store_dirs=[serve_dir])
                self.s = server.Server(
                    'localhost', 32483, server.SeismosizerHandler, self.engine)

            def run(self):
                asyncore.loop(1., use_poll=True)

        t_ws = ServerThread(self.serve_dir)
        t_ws.start()

        try:
            ws.download_gf_store(
                site='http://localhost:32483',
                store_id=self.store_id,
                quiet=False)
            gfstore = store.Store(self.store_id)
            gfstore.check()

        finally:
            # import time
            # time.sleep(100)
            t_ws.s.close()
            t_ws.join(1.)


if __name__ == '__main__':
    util.setup_logging('test_gf_ws', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
import sys
import random  # noqa
import math
import unittest
from tempfile import mkdtemp
import logging
import shutil

import numpy as num

from pyrocko import gf, util
from pyrocko.fomosto import ahfullgreen


logger = logging.getLogger('pyrocko.test.test_gf_ahfull')


r2d = 180. / math.pi
d2r = 1.0 / r2d
km = 1000.


def numeq(a, b, eps):
    return (num.all(num.asarray(a).shape == num.asarray(b).shape and
            num.abs(num.asarray(a) - num.asarray(b)) < eps))


class GFAhfullTestCase(unittest.TestCase):

    tempdirs = []
    if sys.version_info < (2, 7):
        from contextlib import contextmanager

        @contextmanager
        def assertRaises(self, exc):

            gotit = False
            try:
                yield None
            except exc:
                gotit = True

            assert gotit, 'expected to get a %s exception' % exc

        def assertIsNone(self, value):
            assert value is None, 'expected None but got %s' % value

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def test_create_default(self):
        d = mkdtemp(prefix='gfstore')
        self.tempdirs.append(d)
        ahfullgreen.init(d, None)
        store = gf.Store(d)
        store.make_ttt()
        ahfullgreen.build(d)


if __name__ == '__main__':
    util.setup_logging('test_gf_ahfull', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import range
from builtins import map

import random
import time
import tempfile
import os
import fcntl
import unittest
import errno
from pyrocko import util
from pyrocko.parimap import parimap


class Crash(Exception):
    pass


def imapemulation(function, *iterables):
    iterables = list(map(iter, iterables))
    while True:
        args = []
        for it in iterables:
            try:
                args.append(next(it))
            except StopIteration:
                return

        if function is None:
            yield tuple(args)
        else:
            yield function(*args)


class ParimapTestCase(unittest.TestCase):

    def test_parimap(self):

        # random.seed(0)

        for i in range(50):
            nprocs = random.randint(1, 10)
            nx = random.randint(0, 1000)
            ny = random.randint(0, 1000)
            icrash = random.randint(0, 1000)

            # print 'testing %i %i %i %i...' % (nprocs, nx, ny, icrash)

            def work(x, y):
                if x == icrash:
                    raise Crash(str((x, y)))

                a = random.random()
                if a > 0.5:
                    time.sleep((a*0.01)**2)

                return x+y

            I1 = parimap(
                work, range(nx), range(ny),
                nprocs=nprocs, eprintignore=Crash)

            I2 = imapemulation(work, range(nx), range(ny))

            while True:

                exc1, exc2 = None, None
                res1, res2 = None, None
                end1, end2 = None, None
                try:
                    res1 = next(I1)
                except StopIteration:
                    end1 = True
                except Crash as e1:
                    exc1 = e1

                try:
                    res2 = next(I2)
                except StopIteration:
                    end2 = True
                except Crash as e2:
                    exc2 = e2

                assert res1 == res2, str((res1, res2))
                assert type(exc1) == type(exc2)
                assert end1 == end2

                if end1 or end2:
                    break

    def test_locks(self):

        def work(x):
            assert os.path.exists(fn)
            with open(fn, 'a+') as f:
                while True:
                    try:
                        fcntl.lockf(f, fcntl.LOCK_EX)
                        break
                    except IOError as e:
                        if e.errno == errno.ENOLCK:
                            time.sleep(0.01)
                            pass
                        else:
                            raise

                f.seek(0)
                assert '' == f.read()
                f.write('%s' % x)
                f.flush()
                # time.sleep(0.01)
                f.seek(0)
                f.truncate(0)
                fcntl.lockf(f, fcntl.LOCK_UN)

        fos, fn = tempfile.mkstemp()  # (dir='/try/with/nfs/mounted/dir')
        f = open(fn, 'w')
        f.close()

        for x in parimap(work, range(100), nprocs=10, eprintignore=()):
            pass

        os.close(fos)
        os.remove(fn)


if __name__ == '__main__':
    util.setup_logging('test_parimap', 'warning')
    unittest.main()



from __future__ import division
from builtins import range

import math
import hashlib
import logging

import numpy as num

from pyrocko.guts import Object, Int, Bool, Float
from pyrocko import orthodrome as od
from pyrocko.dataset import gshhg, topo

logger = logging.getLogger('pyrocko.scenario.base')

guts_prefix = 'pf.scenario'

km = 1000.
d2r = num.pi/180.
N = 10000000

coastlines = None


def get_gsshg():
    global coastlines
    if coastlines is None:
        logger.debug('Initialising GSHHG database.')
        coastlines = gshhg.GSHHG.low()
    return coastlines


class ScenarioError(Exception):
    pass


def is_on_land(lat, lon, method='coastlines'):
    if method == 'topo':
        elevation = topo.elevation(lat, lon)
        if elevation is None:
            return False
        else:
            return topo.elevation(lat, lon) > 0.

    elif method == 'coastlines':
        logger.debug('Testing %.4f %.4f' % (lat, lon))
        return get_gsshg().is_point_on_land(lat, lon)


def random_lat(rstate, lat_min=-90., lat_max=90.):
    lat_min_ = 0.5*(math.sin(lat_min * math.pi/180.)+1.)
    lat_max_ = 0.5*(math.sin(lat_max * math.pi/180.)+1.)
    return math.asin(rstate.uniform(lat_min_, lat_max_)*2.-1.)*180./math.pi


class Generator(Object):
    seed = Int.T(
        optional=True,
        help='Random seed for a reproducible scenario.')

    def __init__(self, **kwargs):
        Object.__init__(self, **kwargs)
        self._seed = None
        self._parent = None
        self.update_hierarchy()
        self._retry_offset = 0

    def retry(self):
        self.clear()
        self._retry_offset += 1
        for val in self.T.ivals(self):
            if isinstance(val, Generator):
                val.retry()

    def clear(self):
        self._seed = None

    def hash(self):
        return hashlib.sha1(
            (self.dump() + '\n\n%i' % self._retry_offset).encode('utf8'))\
            .hexdigest()

    def get_seed_offset(self):
        return int(self.hash(), base=16) % N

    def update_hierarchy(self, parent=None):
        self._parent = parent
        for val in self.T.ivals(self):
            if isinstance(val, Generator):
                val.update_hierarchy(parent=self)

    def get_seed(self):
        if self._seed is None:
            if self.seed is None:
                if self._parent is not None:
                    self._seed = self._parent.get_seed()
                else:
                    self._seed = num.random.randint(N)
            elif self.seed == 0:
                self._seed = num.random.randint(N)
            else:
                self._seed = self.seed

        return self._seed + self.get_seed_offset()

    def get_rstate(self, i):
        return num.random.RandomState(int(self.get_seed() + i))

    def get_center_latlon(self):
        return self._parent.get_center_latlon()

    def get_radius(self):
        return self._parent.get_radius()

    def get_stations(self):
        return []


class LocationGenerator(Generator):

    avoid_water = Bool.T(
        default=True,
        help='Set whether wet areas should be avoided.')
    center_lat = Float.T(
        optional=True,
        help='Center latitude for the random locations in [deg].')
    center_lon = Float.T(
        optional=True,
        help='Center longitude for the random locations in [deg].')
    radius = Float.T(
        optional=True,
        help='Radius for distribution of random locations [m].')
    ntries = Int.T(
        default=500,
        help='Maximum number of tries to find a location satisifying all '
             'given constraints')

    def __init__(self, **kwargs):
        Generator.__init__(self, **kwargs)
        self._center_latlon = None

    def clear(self):
        Generator.clear(self)
        self._center_latlon = None

    def get_center_latlon(self):
        assert (self.center_lat is None) == (self.center_lon is None)

        if self._center_latlon is None:

            if self.center_lat is not None and self.center_lon is not None:
                self._center_latlon = self.center_lat, self.center_lon

            else:
                if self._parent:
                    self._center_latlon = self._parent.get_center_latlon()
                else:
                    rstate = self.get_rstate(0)
                    lat = random_lat(rstate)
                    lon = rstate.uniform(-180., 180.)
                    self._center_latlon = lat, lon

        return self._center_latlon

    def get_radius(self):
        if self.radius is not None:
            return self.radius
        elif self._parent is not None:
            return self._parent.get_radius()
        else:
            return None

    def get_latlon(self, i):
        rstate = self.get_rstate(i)
        for itry in range(self.ntries):
            radius = self.get_radius()
            if radius is None:
                lat = random_lat(rstate)
                lon = rstate.uniform(-180., 180.)
            else:
                lat_center, lon_center = self.get_center_latlon()
                while True:
                    north = rstate.uniform(-radius, radius)
                    east = rstate.uniform(-radius, radius)
                    if math.sqrt(north**2 + east**2) <= radius:
                        break

                lat, lon = od.ne_to_latlon(lat_center, lon_center, north, east)

            if not self.avoid_water or is_on_land(lat, lon):
                return lat, lon

        if self.avoid_water:
            sadd = ' (avoiding water)'

        raise ScenarioError('could not generate location%s' % sadd)


class TargetGenerator(LocationGenerator):
    pass



import unittest
from pyrocko import util


class BackwardCompatTest(unittest.TestCase):
    pass


compat_modules = [
    'automap',
    'css',
    'datacube',
    'gcf',
    'gse1',
    'gse2_io_wrap',
    'ims',
    'io_common',
    'kan',
    'mseed',
    'rdseed',
    'sacio',
    'segy',
    'seisan_response',
    'seisan_waveform',
    'suds',
    'yaff',
    'catalog',
    'iris_ws',
    'crust2x2',
    'crustdb',
    'geonames',
    'tectonics',
    'topo',
    'automap',
    'beachball',
    'cake_plot',
    'gmtpy',
    'hudson',
    'response_plot',
    'fdsn.__init__',
    'fdsn.enhanced_sacpz',
    'fdsn.station',
    'fdsn.resp',
    'fdsn.ws',
    'marker',
]


def _make_function(module):
    def f(self):
        __import__('pyrocko.' + module)

    f.__name__ = 'test_import_%s' % module

    return f


for mod in compat_modules:
    setattr(
        BackwardCompatTest,
        'test_import_' + mod,
        _make_function(mod))


if __name__ == '__main__':
    util.setup_logging('test_tutorials', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import
from builtins import zip
from builtins import range
from pyrocko import util
import re
import unittest
import tempfile
import shutil
import time
import os
from random import random
import numpy as num


class UtilTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.tempdir = tempfile.mkdtemp()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tempdir)

    def fpath(self, fn):
        return os.path.join(self.tempdir, fn)

    def testTime(self):

        for fmt, accu in zip(
                ['%Y-%m-%d %H:%M:%S.3FRAC', '%Y-%m-%d %H:%M:%S.2FRAC',
                 '%Y-%m-%d %H:%M:%S.1FRAC', '%Y-%m-%d %H:%M:%S',
                 '%Y-%m-%d %H.%M.%S.3FRAC'],
                [0.001, 0.01, 0.1, 1., 0.001, 0.001]):

            ta = util.str_to_time('1960-01-01 10:10:10')
            tb = util.str_to_time('2020-01-01 10:10:10')

            for i in range(10000):
                t1 = ta + random() * (tb-ta)
                s = util.time_to_str(t1, format=fmt)
                t2 = util.str_to_time(s, format=fmt)
                assert abs(t1 - t2) < accu
                fmt_opt = re.sub(r'\.[0-9]FRAC$', '', fmt) + '.OPTFRAC'
                t3 = util.str_to_time(s, format=fmt_opt)
                assert abs(t1 - t3) < accu

    def testBigTime(self):
        s = '2500-01-01 00:00:00.000'
        tx = util.str_to_time(s)
        assert s == util.time_to_str(tx)

    def testIterTimes(self):

        tmin = util.str_to_time('1999-03-20 20:10:10')
        tmax = util.str_to_time('2001-05-20 10:00:05')

        ii = 0
        for ymin, ymax in util.iter_years(tmin, tmax):
            for mmin, mmax in util.iter_months(ymin, ymax):
                ii += 1
                s1 = util.time_to_str(mmin)
                s2 = util.time_to_str(mmax)

        assert ii == 12*3
        assert s1 == '2001-12-01 00:00:00.000'
        assert s2 == '2002-01-01 00:00:00.000'

    def testTimeError(self):
        ok = False
        try:
            util.str_to_time('abc')
        except util.TimeStrError:
            ok = True

        assert ok

    def benchmark_stt_tts(self):
        for x in range(2):
            if x == 1:
                util.util_ext = None
            t = util.str_to_time('1999-03-20 20:10:10')
            tt1 = time.time()
            for i in range(10000):
                s = util.tts(t)
                util.stt(s)

            tt2 = time.time()
            print(tt2 - tt1)

    def test_consistency_merge(self):
        data = [
            ('a', 1, 2, 3.),
            ('a', 2, 2, 3.),
            ('a', 1, 2, 3.)]

        merged = util.consistency_merge(data, error='ignore')
        assert merged == (1, 2, 3.0)

    def test_leap_seconds(self):
        from_sys = {}
        import platform
        if platform.system() != 'Darwin':
            for t, n in util.read_leap_seconds():  # not available on Mac OS X
                from_sys[t] = n

        for t, n in util.read_leap_seconds2():
            if t in from_sys:
                assert from_sys[t] == n

    def test_gps_utc_offset(self):
        for t_utc_0 in [x[0] for x in util.read_leap_seconds2()]:
            t_utc_0 = float(t_utc_0)
            ts_utc = num.linspace(
                t_utc_0 - 2.0, t_utc_0 + 2.0, 17)

            for t_utc in ts_utc:
                t_gps = t_utc + util.gps_utc_offset(t_utc)
                t_utc2 = t_gps + util.utc_gps_offset(t_gps)

                self.assertEqual(util.tts(t_utc), util.tts(t_utc2))

            ts_gps = num.linspace(
                ts_utc[0] + util.gps_utc_offset(ts_utc[0]),
                ts_utc[-1] + util.gps_utc_offset(ts_utc[-1]), 17 + 4)

            t_utc_wrapped = []
            for t_gps in ts_gps:
                t_utc = t_gps + util.utc_gps_offset(t_gps)
                t_utc_wrapped.append(t_utc - t_utc_0)

            num.testing.assert_almost_equal(
                t_utc_wrapped,
                num.concatenate((
                    num.linspace(-2.0, 0.75, 12),
                    num.linspace(0.0, 2.0, 9))))

    def test_plf_integration(self):
        import numpy as num

        x = num.array([1., 1., 3., 3.])
        y = num.array([0., 1., 1., 0.])
        x_edges = num.array([0.5, 1.5, 2.5, 3.5])
        yy = util.plf_integrate_piecewise(x_edges, x, y)
        assert num.all(num.abs(yy - num.array([0.5, 1.0, 0.5])) < 1e-6)

        x = num.array([0., 1., 2., 3.])
        y = num.array([0., 1., 1., 0.])
        x_edges = num.array([0., 1., 2., 3.])
        yy = util.plf_integrate_piecewise(x_edges, x, y)
        assert num.all(num.abs(yy - num.array([0.5, 1.0, 0.5])) < 1e-6)

    def test_arange2(self):
        num.testing.assert_almost_equal(
            util.arange2(0., 1., 0.1), num.linspace(0., 1., 11))

        with self.assertRaises(util.ArangeError):
            util.arange2(0., 1.05, 0.1)

        num.testing.assert_almost_equal(
            util.arange2(0., 1.04, 0.1, error='round'),
            num.linspace(0., 1., 11))

        num.testing.assert_almost_equal(
            util.arange2(0., 1.05, 0.1, error='floor'),
            num.linspace(0., 1., 11))

        num.testing.assert_almost_equal(
            util.arange2(0., 1.05, 0.1, error='ceil'),
            num.linspace(0., 1.1, 12))

    def test_gform(self):
        s = ''
        for i in range(-11, 12):
            v = 1/3. * 10**i
            s += '|%s|\n' % util.gform(v)

        self.assertEqual(s.strip(), '''
|   3.33E-12 |
|   3.33E-11 |
|   3.33E-10 |
|   3.33E-09 |
|   3.33E-08 |
|   3.33E-07 |
|   3.33E-06 |
|   3.33E-05 |
|   3.33E-04 |
|   3.33E-03 |
|   3.33E-02 |
|   0.333    |
|   3.33     |
|  33.3      |
| 333.       |
|   3.33E+03 |
|   3.33E+04 |
|   3.33E+05 |
|   3.33E+06 |
|   3.33E+07 |
|   3.33E+08 |
|   3.33E+09 |
|   3.33E+10 |'''.strip())

    def test_download(self):
        fn = self.fpath('responses.xml')
        url = 'http://data.pyrocko.org/examples/responses.xml'

        stat = []

        def status(d):
            stat.append(d)

        util.download_file(url, fn, status_callback=status)

        url = 'http://data.pyrocko.org/testing/my_test_dir'
        dn = self.fpath('my_test_dir')
        util.download_dir(url, dn, status_callback=status)

        d = stat[-1]

        dwant = {
            'ntotal_files': 4,
            'nread_files': 4,
            'ntotal_bytes_all_files': 22,
            'nread_bytes_all_files': 22,
            'ntotal_bytes_current_file': 8,
            'nread_bytes_current_file': 8}

        for k in dwant:
            assert k in d
            assert d[k] == dwant[k]


if __name__ == "__main__":
    util.setup_logging('test_util', 'info')
    unittest.main()



# python 2/3
from __future__ import division, print_function, absolute_import

import unittest
from pyrocko.dataset import geonames
from pyrocko import util


class GeonamesTestCase(unittest.TestCase):

    def test_geonames(self):

        cities = geonames.get_cities(53.6, 10.0, 100e3, 200000)
        self.assertEqual(
            sorted(c.asciiname for c in cities),
            ['Bremen', 'Hamburg', 'Kiel', 'Luebeck'])


if __name__ == "__main__":
    util.setup_logging('test_geonames', 'warning')
    unittest.main()



from __future__ import division, print_function, absolute_import

import unittest
import logging
from tempfile import mkdtemp
import shutil

from pyrocko import util, trace, gf, cake  # noqa
from pyrocko.fomosto import qseis
from pyrocko.fomosto.report import GreensFunctionTest as gftest

logger = logging.getLogger('pyrocko.test.test_fomosto_report')

km = 1e3


class ReportTestCase(unittest.TestCase):
    tempdirs = []

    def __init__(self, *args, **kwargs):
        unittest.TestCase.__init__(self, *args, **kwargs)

    @classmethod
    def tearDownClass(cls):
        for d in cls.tempdirs:
            shutil.rmtree(d)

    def test_pyrocko_report_with_gf_and_qseis(self):

        mod = cake.LayeredModel.from_scanlines(cake.read_nd_model_str('''
 0. 5.8 3.46 2.6 1264. 600.
 20. 5.8 3.46 2.6 1264. 600.
 20. 6.5 3.85 2.9 1283. 600.
 35. 6.5 3.85 2.9 1283. 600.
mantle
 35. 8.04 4.48 3.58 1449. 600.
 77.5 8.045 4.49 3.5 1445. 600.
 77.5 8.045 4.49 3.5 180.6 75.
 120. 8.05 4.5 3.427 180. 75.
 120. 8.05 4.5 3.427 182.6 76.06
 165. 8.175 4.509 3.371 188.7 76.55
 210. 8.301 4.518 3.324 201. 79.4
 210. 8.3 4.52 3.321 336.9 133.3
 410. 9.03 4.871 3.504 376.5 146.1
 410. 9.36 5.08 3.929 414.1 162.7
 660. 10.2 5.611 3.918 428.5 172.9
 660. 10.79 5.965 4.229 1349. 549.6'''.lstrip()))

        store_dir = mkdtemp(prefix='gft_')
        self.tempdirs.append(store_dir)

        qsconf = qseis.QSeisConfig()
        qsconf.qseis_version = '2006a'

        qsconf.time_region = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.cut = (
            gf.meta.Timing('0'),
            gf.meta.Timing('end+100'))

        qsconf.wavelet_duration_samples = 0.001
        qsconf.sw_flat_earth_transform = 0

        config = gf.meta.ConfigTypeA(
            id='gft_test',
            ncomponents=10,
            sample_rate=0.25,
            receiver_depth=0.*km,
            source_depth_min=10*km,
            source_depth_max=10*km,
            source_depth_delta=1*km,
            distance_min=550*km,
            distance_max=560*km,
            distance_delta=1*km,
            modelling_code_id='qseis.2006a',
            earthmodel_1d=mod,
            tabulated_phases=[
                gf.meta.TPDef(
                    id='begin',
                    definition='p,P,p\\,P\\'),
                gf.meta.TPDef(
                    id='end',
                    definition='2.5'),
            ])

        config.validate()
        gf.store.Store.create_editables(
            store_dir, config=config, extra={'qseis': qsconf})

        store = gf.store.Store(store_dir, 'r')
        store.make_ttt()
        store.close()

        try:
            qseis.build(store_dir, nworkers=1)
        except qseis.QSeisError as e:
            if str(e).find('could not start qseis') != -1:
                logger.warn('qseis not installed; '
                            'skipping test_pyrocko_gf_vs_qseis')
                return
            else:
                raise

        gft = gftest(store_dir, sensor_count=2, pdf_dir=store_dir,
                     plot_velocity=True, rel_lowpass_frequency=(1. / 110),
                     rel_highpass_frequency=(1. / 16), output_format='html')

        src = gft.createSource('DC', None, 45., 90., 180.)
        sen = gft.createSensors(strike=0., codes=('', 'STA', '', 'Z'),
                                azimuth=0., dip=-90.)
        gft.trace_configs = [[src, sen]]
        gft.createDisplacementTraces()
        gft.createVelocityTraces()
        gft.applyFrequencyFilters()
        gft.getPhaseArrivals()
        gft.createOutputDoc()


if __name__ == '__main__':
    util.setup_logging('test_fomosto_report', 'warning')
    unittest.main()
